
def test_RandomConstraint():
    """ Test the RandomConstraint class in acquisitions.py.

    Use the RandomConstraint class to randomly select a target point and
    improvement direction, and check that exactly one point was targeted.

    """

    from jax import jacrev
    import numpy as np
    from parmoo.acquisitions import RandomConstraint
    from parmoo.util import updatePF
    import pytest

    # Define the objective function
    def obj_f(x):
        return np.asarray([np.dot(x - np.eye(3)[i],
                                  x - np.eye(3)[i]) for i in range(3)])

    # Generate a database
    x_vals = np.random.random_sample((10, 3))
    data = {'x_vals': x_vals,
            'f_vals': np.asarray([obj_f(x) for x in x_vals]),
            'c_vals': np.ones((10, 1))}

    # Initialize a good instance
    acqu = RandomConstraint(3, np.zeros(3), np.ones(3), {})
    # Try some bad targets to test error handling
    with pytest.raises(TypeError):
        acqu.setTarget(5, lambda x: np.zeros(1))
    with pytest.raises(AttributeError):
        acqu.setTarget({'x_vals': []}, lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones(1), 'f_vals': np.ones(2)},
                       lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 1)), 'f_vals': np.ones((1, 3))},
                       lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 3)), 'f_vals': np.ones((1, 1))},
                       lambda x: np.zeros(0))
    with pytest.raises(TypeError):
        acqu.setTarget(data, 5)
    with pytest.raises(ValueError):
        acqu.setTarget(data, lambda x, y, z: np.zeros(0))
    data['c_vals'] = np.zeros((10, 1))
    # Set a few good target
    assert (np.all(acqu.setTarget({}, lambda x: np.zeros(3)) < 1.0))
    assert (np.all(acqu.setTarget({}, lambda x: np.zeros(3)) > 0.0))
    assert (np.all(acqu.setTarget({'x_vals': np.zeros((1, 3)),
                                   'f_vals': np.zeros((1, 3)),
                                   'c_vals': np.zeros((1, 1))},
                                  lambda x: np.ones(3) * (0.01 - sum(x)))
                   < 1.0001))
    assert (acqu.setTarget(data, lambda x: np.zeros(3)) in data['x_vals'])
    # Generate a random scalarization target and check the scalarization
    acqu = RandomConstraint(3, np.zeros(3), np.ones(3), {})
    acqu.setTarget({'x_vals': None, 'f_vals': None},
                   lambda x: np.zeros(3))
    acqu.setTarget(data, lambda x: np.zeros(3))
    # Get a copy of the Pareto front for checking correctness
    pf = updatePF(data, {})
    # Check that the scalar value is either less than the sum of fi or bad
    for fi in pf['f_vals']:
        assert (acqu.scalarize(fi, np.zeros(3), np.zeros(3), np.zeros(3))
                <= np.sum(fi) + 1.0e-4 or np.any(fi > acqu.f_ub - 1.0e-4))
    # Check that the scalar grad is either less than sum of weights or bad
    grad = jacrev(acqu.scalarize, argnums=0)
    for fi in pf['f_vals']:
        xi, si, sdi = np.zeros(3), np.zeros(1), np.zeros(1)
        assert (np.all(grad(fi, xi, si, sdi) <= np.sum(acqu.weights) + 1.0e-4)
                or np.any(fi > acqu.f_ub - 1.0e-4))


def test_EI_RandomConstraint():
    """ Test the EI_RandomConstraint class in acquisitions.py.

    Use the EI_RandomConstraint class to randomly select a target point and
    improvement direction, and check that exactly one point was targeted.

    """

    from parmoo.acquisitions import EI_RandomConstraint
    from parmoo.util import updatePF
    import numpy as np
    import pytest

    # Define the objective function
    def obj_f(x):
        return np.asarray([np.dot(x - np.eye(3)[i],
                                  x - np.eye(3)[i]) for i in range(3)])

    # Generate a database
    x_vals = np.random.random_sample((10, 3))
    data = {'x_vals': x_vals,
            'f_vals': np.asarray([obj_f(x) for x in x_vals]),
            'c_vals': np.ones((10, 1))}

    # Initialize a good instance
    acqu = EI_RandomConstraint(3, np.zeros(3), np.ones(3), {})
    # Try some bad targets to test error handling
    with pytest.raises(TypeError):
        acqu.setTarget(5, lambda x: np.zeros(1))
    with pytest.raises(AttributeError):
        acqu.setTarget({'x_vals': []}, lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones(1), 'f_vals': np.ones(2)},
                       lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 1)), 'f_vals': np.ones((1, 3))},
                       lambda x: np.zeros(0))
    with pytest.raises(ValueError):
        acqu.setTarget({'x_vals': np.ones((1, 3)), 'f_vals': np.ones((1, 1))},
                       lambda x: np.zeros(0))
    with pytest.raises(TypeError):
        acqu.setTarget(data, 5)
    with pytest.raises(ValueError):
        acqu.setTarget(data, lambda x, y, z: np.zeros(0))
    data['c_vals'] = np.zeros((10, 1))
    # Set a few good target
    assert (np.all(acqu.setTarget({}, lambda x, sx=0: np.zeros(3)) < 1.0))
    assert (np.all(acqu.setTarget({}, lambda x, sx=0: np.zeros(3)) > 0.0))
    assert (np.all(acqu.setTarget({'x_vals': np.zeros((1, 3)),
                                   'f_vals': np.zeros((1, 3)),
                                   'c_vals': np.zeros((1, 1))},
                                  lambda x, sx=0: np.ones(3) * (0.01 - sum(x)))
                   < 1.0001))
    assert (acqu.setTarget(data, lambda x, sx=0: np.zeros(3))
            in data['x_vals'])
    # Generate a random 1D scalarization target and check the scalarization
    acqu = EI_RandomConstraint(3, np.zeros(3), np.ones(3), {})
    acqu.setTarget({'x_vals': None, 'f_vals': None},
                   lambda x, sx=0: np.zeros(3))
    acqu.setTarget(data, lambda x, sx=0: np.zeros(3))
    # Get a copy of the Pareto front for checking correctness
    pf = updatePF(data, {})
    # Check that the scalar value is either less than the sum of fi or bad
    for fi in pf['f_vals']:
        assert (acqu.scalarize(fi, np.zeros(3), np.zeros(1), np.ones(1))
                <= np.sum(fi) + 1.0e-4 or np.any(fi > acqu.f_ub - 1.0e-4))
    # Generate a random 2D scalarization target and check the scalarization
    acqu = EI_RandomConstraint(3, np.zeros(3), np.ones(3), {})
    acqu.setTarget({'x_vals': None, 'f_vals': None},
                   lambda x, sx=0: np.zeros(3))
    acqu.setTarget(data, lambda x, sx=0: np.zeros(3))
    # Get a copy of the Pareto front for checking correctness
    pf = updatePF(data, {})
    # Check that the scalar value is either less than the sum of fi or bad
    for fi in pf['f_vals']:
        assert (acqu.scalarize(fi, np.zeros(3), np.zeros(2), np.ones(2))
                <= np.sum(fi) + 1.0e-4 or np.any(fi > acqu.f_ub - 1.0e-4))
    # Generate a random 3D scalarization target and check the scalarization
    acqu = EI_RandomConstraint(3, np.zeros(3), np.ones(3), {})
    acqu.setTarget({'x_vals': None, 'f_vals': None},
                   lambda x, sx=0: np.zeros(3))
    acqu.setTarget(data, lambda x, sx=0: np.zeros(3))
    # Get a copy of the Pareto front for checking correctness
    pf = updatePF(data, {})
    # Check that the scalar value is either less than the sum of fi or bad
    for fi in pf['f_vals']:
        assert (acqu.scalarize(fi, np.zeros(3), np.zeros(3), np.ones(3))
                <= np.sum(fi) + 1.0e-4 or np.any(fi > acqu.f_ub - 1.0e-4))


if __name__ == "__main__":
    test_RandomConstraint()
    test_EI_RandomConstraint()




import numpy as np
from parmoo import MOOP
from parmoo.searches import LatinHypercube
from parmoo.surrogates import GaussRBF
from parmoo.acquisitions import RandomConstraint
from parmoo.optimizers import GlobalSurrogate_PS

# Fix the random seed for reproducibility using the np_random_gen hyperparams
my_moop = MOOP(GlobalSurrogate_PS, hyperparams={'np_random_gen': 0})

my_moop.addDesign({'name': "x1",
                   'des_type': "continuous",
                   'lb': 0.0, 'ub': 1.0})
# Note: the 'levels' key can contain a list of strings, but jax can only jit
# numeric types, so integer level IDs are strongly recommended
my_moop.addDesign({'name': "x2", 'des_type': "categorical",
                   'levels': [-1, 1]})

def sim_func(x):
   sx = np.array([(x["x1"] - 0.2) ** 2, (x["x1"] - 0.8) ** 2])
   ## The following 2 lines are equivalent, but jax cannot jit if statements.
   ## Uncomment below to see the difference in execution speed from jit
   # if x["x2"] != 1: sx += 99.
   sx += 99. - 99. * (x["x2"] == 1)
   return sx

my_moop.addSimulation({'name': "MySim",
                       'm': 2,
                       'sim_func': sim_func,
                       'search': LatinHypercube,
                       'surrogate': GaussRBF,
                       'hyperparams': {'search_budget': 20}})

def f1(x, s): return s["MySim"][0]
def f2(x, s): return s["MySim"][1]
my_moop.addObjective({'name': "f1", 'obj_func': f1})
my_moop.addObjective({'name': "f2", 'obj_func': f2})

def c1(x, s): return 0.1 - x["x1"]
my_moop.addConstraint({'name': "c1", 'constraint': c1})

for i in range(3):
   my_moop.addAcquisition({'acquisition': RandomConstraint,
                           'hyperparams': {}})

my_moop.solve(5)
results = my_moop.getPF(format="pandas")

# Display solution
print(results)

# Plot results -- must have extra viz dependencies installed
from parmoo.viz import scatter
# The optional arg `output` exports directly to jpeg instead of interactive mode
scatter(my_moop, output="jpeg")



""" This module contains a library of common objective functions, matching
ParMOO's interface.

The common objectives are:
 * ``SingleSimObjective`` -- min or max a single simulation output
 * ``SumOfSimSquaresObjective`` -- min or max several squared sim outputs
 * ``SumOfSimsObjective`` -- min or max the (absolute) sum of sim outputs

And the corresponding gradients are defined in:
 * ``SingleSimGradient``
 * ``SumOfSimSquaresGradient``
 * ``SumOfSimsGradient``

"""

from jax import numpy as jnp
import numpy as np
from parmoo.structs import CompositeFunction


class SingleSimObjective(CompositeFunction):
    """ Class for min/maxing a single simulation's output.

    If minimizing:

    ``def ObjectiveFunction(x, sx): return sx[self.sim_ind]``

    If maximizing:

    ``def ObjectiveFunction(x, sx): return -sx[self.sim_ind]``

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_ind, goal='min')``
     * ``__call__(x, sim)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` returns sim[self.sim_ind].

    """

    def __init__(self, des_type, sim_type, sim_ind, goal='min'):
        """ Constructor for SingleSimObjective class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_ind (str or tuple): The name (or name, output index pair
                for simulations with more than one output field)
                designating the simulation output to minimize or maximize.

            goal (str): Either 'min' to minimize or 'max' to maximize.
                Defaults to 'min'.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if isinstance(sim_ind, str):
            try:
                assert (sim_ind in np.dtype(self.sim_type).names)
            except BaseException:
                raise ValueError(f"{sim_ind[0]} not a name in given sim_type")
            self.sim_name = sim_ind
            self.goal = jnp.ones(1)
        elif isinstance(sim_ind, tuple):
            try:
                assert sim_ind[0] in np.dtype(self.sim_type).names
            except BaseException:
                raise ValueError(f"{sim_ind[0]} not a name in given sim_type")
            try:
                assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
            except BaseException:
                raise ValueError(f"{sim_ind[1]} not an index of {sim_ind[0]}")
            self.sim_name = sim_ind[0]
            sim_size = self.sim_type[sim_ind[0]].shape[0]
            self.goal = jnp.eye(sim_size)[sim_ind[1]]
        else:
            raise TypeError("sim_ind must be a str or (str, int) tuple")
        # Check for optional input
        if goal.lower() not in ('min', 'max'):
            raise ValueError("goal must be either 'min' or 'max'")
        if goal.lower() == 'max':
            self.goal *= -1.0
        return

    def __call__(self, x, sx):
        """ Define objective evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of the simulation at self.sim_ind given the
            input x.

        """

        # These lines look silly, but do not change them.
        # They are needed to gracefully cast all possible shapes to scalar.
        fx = 0.0
        fx += jnp.dot(sx[self.sim_name], self.goal)
        return fx


class SumOfSimSquaresObjective(CompositeFunction):
    """ Class for min/maxing the sum-of-squared simulation outputs.

    If minimizing, equivalent to:

    ```
    def ObjectiveFunction(x, sx):
        return sum([sx[i]**2 for i in sim_inds])
    ```

    If maximizing, equivalent to:

    ```
    def ObjectiveFunction(x, sx):
        return -sum([sx[i]**2 for i in sim_inds])
    ```

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_inds, goal='min')``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` evaluates the sum-of-squared simulation outputs.

    """

    def __init__(self, des_type, sim_type, sim_inds, goal='min'):
        """ Constructor for SumOfSimSquaresObjective class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_inds (list): The list of names (or name, index pairs for sims
                with more than one output) of the simulation outputs to sum
                over.

            goal (str): Either 'min' to minimize SOS or 'max' to maximize SOS.
                Defaults to 'min'.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if not isinstance(sim_inds, list):
            raise TypeError("sim_inds must be a list of ints, tuples, or " +
                            "strings")
        self.sim_names = []
        self.sim_inds = []
        for sim_ind in sim_inds:
            if isinstance(sim_ind, str):
                try:
                    assert (sim_ind in np.dtype(self.sim_type).names)
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                if sim_ind in self.sim_names:
                    self.sim_inds[self.sim_inds.index(sim_ind)] += 1.0
                else:
                    self.sim_names.append(sim_ind)
                    self.sim_inds.append(jnp.ones(1))
            elif isinstance(sim_ind, tuple):
                try:
                    assert sim_ind[0] in np.dtype(self.sim_type).names
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                try:
                    assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
                except BaseException:
                    raise ValueError(f"{sim_ind[1]} not an index of "
                                     f"{sim_ind[0]}")
                sim_size = self.sim_type[sim_ind[0]].shape[0]
                if sim_ind[0] in self.sim_names:
                    self.sim_inds[self.sim_names.index(sim_ind[0])] += \
                            jnp.eye(sim_size)[sim_ind[1]]
                else:
                    self.sim_names.append(sim_ind[0])
                    self.sim_inds.append(jnp.eye(sim_size)[sim_ind[1]])
            else:
                raise TypeError("Each sim_ind must be a str or str,int tuple")
        # Check for optional input
        if goal.lower() not in ('min', 'max'):
            raise ValueError("goal must be either 'min' or 'max'")
        if goal.lower() == 'max':
            self.goal = -1.0
        else:
            self.goal = 1.0
        return

    def __call__(self, x, sx):
        """ Define objective evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The sum of squared outputs of the simulation at indices in
            self.sim_ind given the input x.

        """

        fx = 0.0
        for sn, si in zip(self.sim_names, self.sim_inds):
            fx += jnp.dot(sx[sn] ** 2, si)
        return fx * self.goal


class SumOfSimsObjective(CompositeFunction):
    """ Class for min/maxing the sum of simulation outputs.

    If minimizing:

    ``def ObjectiveFunction(x, sx): return sum([sx[i] for i in sim_inds])``

    If maximizing:

    ``def ObjectiveFunction(x, sx): return -sum([sx[i] for i in sim_inds])``

    If minimizing absolute sum:

    ```
    def ObjectiveFunction(x, sx): return sum([abs(sx[i]) for i in sim_inds])
    ```

    If maximizing absolute sum:

    ```
    def ObjectiveFunction(x, sx): return -sum([abs(sx[i]) for i in sim_inds])
    ```

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_inds, goal='min', absolute=False)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` evaluate the (absolute) sum outputs.

    """

    def __init__(self, des_type, sim_type, sim_inds, goal='min',
                 absolute=False):
        """ Constructor for SumOfSimsObjective class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_inds (list): The list of indices or names of the
                simulation outputs to sum over.

            goal (str): Either 'min' to minimize sum or 'max' to maximize sum.
                Defaults to 'min'.

            absolute (bool): True to min/max absolute sum, False to
                min/max raw sum. Defaults to False.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if not isinstance(sim_inds, list):
            raise TypeError("sim_inds must be a list of ints, tuples, or " +
                            "strings")
        self.sim_names = []
        self.sim_inds = []
        for sim_ind in sim_inds:
            if isinstance(sim_ind, str):
                try:
                    assert (sim_ind in np.dtype(self.sim_type).names)
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                if sim_ind in self.sim_names:
                    self.sim_inds[self.sim_inds.index(sim_ind)] += 1.0
                else:
                    self.sim_names.append(sim_ind)
                    self.sim_inds.append(jnp.ones(1))
            elif isinstance(sim_ind, tuple):
                try:
                    assert sim_ind[0] in np.dtype(self.sim_type).names
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                try:
                    assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
                except BaseException:
                    raise ValueError(f"{sim_ind[1]} not an index of "
                                     f"{sim_ind[0]}")
                sim_size = self.sim_type[sim_ind[0]].shape[0]
                if sim_ind[0] in self.sim_names:
                    self.sim_inds[self.sim_names.index(sim_ind[0])] += \
                            jnp.eye(sim_size)[sim_ind[1]]
                else:
                    self.sim_names.append(sim_ind[0])
                    self.sim_inds.append(jnp.eye(sim_size)[sim_ind[1]])
            else:
                raise TypeError("Each sim_ind must be a str or str,int tuple")
        # Check for optional input
        if goal.lower() not in ('min', 'max'):
            raise ValueError("goal must be either 'min' or 'max'")
        if goal.lower() == 'max':
            self.goal = -1.0
        else:
            self.goal = 1.0

        def id_func(x): return np.ones(x.size)

        def abs_func(x): return jnp.sign(x)

        if not isinstance(absolute, bool):
            raise TypeError("absolute must be a bool type, not " +
                            str(type(absolute)))
        if absolute:
            self.absolute = abs_func
        else:
            self.absolute = id_func
        return

    def __call__(self, x, sx):
        """ Define objective evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The sum of the simulation outputs at indices in
            self.sim_ind given the input x.

        """

        fx = 0.0
        for sn, si in zip(self.sim_names, self.sim_inds):
            fx += jnp.sum(self.absolute(sx[sn]) * sx[sn] * si)
        return fx * self.goal


class SingleSimGradient(SingleSimObjective):
    """ Gradient class for SingleSimObjective.

    Inherits from the ``SingleSimObjective`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient output of the simulation at self.sim_ind
            given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        ds[self.sim_name] = self.goal
        return dx, ds


class SumOfSimSquaresGradient(SumOfSimSquaresObjective):
    """ Gradient class for SumOfSimSquaresObjective.

    Inherits from the ``SumOfSimSquaresObjective`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of the sum of squared simulation outputs
            given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        for sn, si in zip(self.sim_names, self.sim_inds):
            ds[sn] = sx[sn] * si * 2.0 * self.goal
        return dx, ds


class SumOfSimsGradient(SumOfSimsObjective):
    """ Gradient class for SumOfSimsObjective.

    Inherits from the ``SumOfSimsObjective`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of the sum of (absolute) simulation
            outputs given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        for sn, si in zip(self.sim_names, self.sim_inds):
            ds[sn] = self.absolute(sx[sn]) * si * self.goal
        return dx, ds




def test_LocalSurrogate_PS():
    """ Test the LocalSurrogate_PS class in optimizers.py.

    Perform a test of the LocalSurrogate_PS class by minimizing the three
    variable, biobjective function

    $$
    F(x) = (-x_1 + x_2 + x_3, x_1 - x_2 + x_3)
    $$

    s.t. $x in [0, 1]^n$.

    Use the weights [1, 0], [0, 1], and [0.5, 0.5].

    Assert that the solutions are all correct. I.e., (1, 0, 0) is
    the minimizer of $F^T [1, 0]$; (0, 1, 0) is the minimizer of $F^T [0, 1]$;
    and the minimizer of $F^T [0.5, 0.5]$ satisfies x_3 = 0.

    """

    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import LocalSurrogate_PS
    import numpy as np
    import pytest

    # Initialize the problem dimensions
    n = 3
    o = 2
    lb = np.zeros(n)
    ub = np.ones(n)
    # Create the biobjective function
    def f(z, sz): return np.asarray([-z[0] + z[1] + z[2], z[0] - z[1] + z[2]])

    def L(z, sz):
        return f(z, sz) + 5 * (max(0.1 - z[2], 0) + max(z[2] - 0.6, 0))

    def S(z): return np.ones(2)

    def SD(z): return np.zeros(2)
    # Create 2 acquisition functions targeting 2 "pure" solutions
    acqu1 = UniformWeights(o, lb, ub, {})
    acqu1.setTarget({}, lambda x: np.zeros(2))
    acqu1.weights[:] = 0.0
    acqu1.weights[0] = 1.0
    acqu2 = UniformWeights(o, lb, ub, {})
    acqu2.setTarget({}, lambda x: np.zeros(2))
    acqu2.weights[:] = 0.0
    acqu2.weights[1] = 1.0
    # Create a third acquisition function targeting a random tradeoff solution
    acqu3 = UniformWeights(o, lb, ub, {})
    acqu3.setTarget({}, lambda x: np.zeros(2))
    acqu3.weights[:] = 0.5
    # Try some bad initializations to test error handling
    with pytest.raises(TypeError):
        LocalSurrogate_PS(o, lb, ub, {'opt_budget': 2.0})
    with pytest.raises(ValueError):
        LocalSurrogate_PS(o, lb, ub, {'opt_budget': 0})
    # Initialize the problem correctly, with and without an optional budget
    LocalSurrogate_PS(o, lb, ub, {'opt_budget': 100})
    opt = LocalSurrogate_PS(o, lb, ub, {})
    # Try to add some bad objectives, constraints, and acquisitions
    with pytest.raises(TypeError):
        opt.setObjective(5)
    with pytest.raises(ValueError):
        opt.setObjective(lambda z1, z2, z3: np.zeros(1))
    with pytest.raises(TypeError):
        opt.setConstraints(5)
    with pytest.raises(ValueError):
        opt.setConstraints(lambda z1, z2, z3: np.zeros(1))
    with pytest.raises(TypeError):
        opt.addAcquisition(5)
    # Try to solve with invalid inputs to test error handling
    with pytest.raises(ValueError):
        opt.solve(np.zeros((3, n-1)))
    with pytest.raises(ValueError):
        opt.solve(np.zeros((4, n)))
    # Set the acquisition and reset functions
    opt.setSimulation(S, SD)
    opt.addAcquisition(acqu1, acqu2, acqu3)
    opt.setTrFunc(lambda x, r: 100.0)
    # Define the solution
    x1_soln = np.eye(n)[0]
    x1_soln[n-1] = 0.1
    x2_soln = np.eye(n)[1]
    x2_soln[n-1] = 0.1
    # Solve the surrogate problem with LocalSurrogate_PS starting from centroid
    x = np.zeros((3, n))
    x[:] = 0.5
    for i in range(10):
        # Add the correct objectives and constraints
        opt.setObjective(f)
        opt.setConstraints(lambda z, sz: np.asarray([0.1 - z[2], z[2] - 0.6]))
        opt.setPenalty(L)
        (x1, x2, x3) = opt.solve(x)
        x[0] = x1
        x[1] = x2
        x[2] = x3
        for j in range(3):
            opt.returnResults(x[j], np.ones(2) * -10, np.zeros(1), np.zeros(1))
    # eps is the tolerance for rejecting a solution as incorrect
    eps = 0.01
    # Check that the computed solutions are within eps of the truth
    assert (np.linalg.norm(x1 - x1_soln) < eps)
    assert (np.linalg.norm(x2 - x2_soln) < eps)
    assert (np.abs(x3[n-1] - 0.1) < eps)
    return


def test_GlobalSurrogate_PS():
    """ Test the GlobalSurrogate_PS class in optimizers.py.

    Perform a test of the globalGPS class by minimizing the three variable,
    biobjective function

    $$
    F(x) = (-x_1 + x_2 + x_3, x_1 - x_2 + x_3)
    $$

    s.t. $x in [0, 1]^n$.

    Use the weights [1, 0], [0, 1], and [0.5, 0.5].

    Assert that the solutions are all correct. I.e., (1, 0, 0) is
    the minimizer of $F^T [1, 0]$; (0, 1, 0) is the minimizer of $F^T [0, 1]$;
    and the minimizer of $F^T [0.5, 0.5]$ satisfies x_3 = 0.

    """

    from parmoo.acquisitions import UniformWeights
    from parmoo.optimizers import GlobalSurrogate_PS
    import numpy as np
    import pytest

    # Initialize the problem dimensions
    n = 3
    o = 2
    lb = np.zeros(n)
    ub = np.ones(n)
    # Create the biobjective function
    def f(z, sz): return np.asarray([-z[0] + z[1] + z[2], z[0] - z[1] + z[2]])

    def L(z, sz):
        return f(z, sz) + 5 * (max(0.1 - z[2], 0) + max(z[2] - 0.6, 0))

    def S(z): return np.ones(2)

    def SD(z): return np.zeros(2)
    # Create 2 acquisition functions targeting 2 "pure" solutions
    acqu1 = UniformWeights(o, lb, ub, {})
    acqu1.setTarget({}, lambda x: np.zeros(2))
    acqu1.weights[:] = 0.0
    acqu1.weights[0] = 1.0
    acqu2 = UniformWeights(o, lb, ub, {})
    acqu2.setTarget({}, lambda x: np.zeros(2))
    acqu2.weights[:] = 0.0
    acqu2.weights[1] = 1.0
    # Create a third acquisition function targeting a random tradeoff solution
    acqu3 = UniformWeights(o, lb, ub, {})
    acqu3.setTarget({}, lambda x: np.zeros(2))
    acqu3.weights[:] = 0.5
    # Try some bad initializations to test error handling
    with pytest.raises(TypeError):
        GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 2.0})
    with pytest.raises(ValueError):
        GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 0})
    with pytest.raises(TypeError):
        GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 500, 'gps_budget': 2.0})
    with pytest.raises(ValueError):
        GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 500, 'gps_budget': 0})
    with pytest.raises(ValueError):
        GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 500, 'gps_budget': 1000})
    # Initialize the problem correctly, with and without an optional budget
    GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 200})
    GlobalSurrogate_PS(o, lb, ub, {'opt_budget': 200, 'gps_budget': 100})
    opt = GlobalSurrogate_PS(o, lb, ub, {})
    # Try to add some bad objectives, constraints, and acquisitions
    with pytest.raises(TypeError):
        opt.setObjective(5)
    with pytest.raises(ValueError):
        opt.setObjective(lambda z1, z2, z3: np.zeros(1))
    with pytest.raises(TypeError):
        opt.setConstraints(5)
    with pytest.raises(ValueError):
        opt.setConstraints(lambda z1, z2, z3: np.zeros(1))
    with pytest.raises(TypeError):
        opt.addAcquisition(5)
    # Try to solve with invalid inputs to test error handling
    with pytest.raises(ValueError):
        opt.solve(np.zeros((3, n-1)))
    with pytest.raises(ValueError):
        opt.solve(np.zeros((4, n)))
    # Add the correct objective and constraints
    opt.setObjective(f)
    opt.setConstraints(lambda z, sz: np.asarray([0.1 - z[2], z[2] - 0.6]))
    opt.setSimulation(S, SD)
    opt.setPenalty(L)
    opt.addAcquisition(acqu1, acqu2, acqu3)
    opt.setTrFunc(lambda x, r: 100.0)
    # Solve the surrogate problem with GlobalSurrogate_PS
    x = np.zeros((3, n))
    x[:, :] = 0.5
    (x1, x2, x3) = opt.solve(x)
    # Define the solution
    x1_soln = np.eye(n)[0]
    x1_soln[n-1] = 0.1
    x2_soln = np.eye(n)[1]
    x2_soln[n-1] = 0.1
    # eps is the tolerance for rejecting a solution as incorrect
    eps = 0.1
    # Check that the computed solutions are within eps of the truth
    assert (np.linalg.norm(x1 - x1_soln) < eps)
    assert (np.linalg.norm(x2 - x2_soln) < eps)
    assert (np.abs(x3[n-1] - 0.1) < eps)
    return


if __name__ == "__main__":
    test_LocalSurrogate_PS()
    test_GlobalSurrogate_PS()



""" This module contains a library of common constraint functions, matching
ParMOO's interface.

The common constraints are:
 * ``SingleSimBound`` -- bound on a single simulation output's value
 * ``SumOfSimSquaresBound`` -- bound on the SOS for several sim outputs
 * ``SumOfSimsBound`` -- bound on the (abs) sum of several sim outputs

And their corresponding gradient functions are:
 * ``SingleSimBoundGradient``
 * ``SumOfSimSquaresBoundGradient``
 * ``SumOfSimsBoundGradient``

"""

from jax import numpy as jnp
import numpy as np
from parmoo.structs import CompositeFunction


class SingleSimBound(CompositeFunction):
    """ Class for bounding a single simulation's output.

    If upper-bounding:

    ```
    def ConstraintFunction(x, sx):
        return sx[self.sim_ind] - upper_bound
    ```

    If lower-bounding:

    ```
    def ConstraintFunction(x, sx):
        return lower_bound - sx[self.sim_ind]
    ```

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_ind, bound_type='min', bound=0)``
     * ``__call__(x, sim)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` returns the slack (negative when feasible).

    """

    def __init__(self, des_type, sim_type, sim_ind,
                 bound_type='upper', bound=0.0):
        """ Constructor for SingleSimBound class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_ind (str or tuple): The name (or name, output index pair
                for simulations with more than one output field)
                designating the simulation output to minimize or maximize.

            bound_type (str): Either 'lower' to lower-bound or 'upper' to
                upper-bound. Defaults to 'upper'.

            bound (float): The lower/upper bound for this constraint.
                Defaults to 0.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if isinstance(sim_ind, str):
            try:
                assert (sim_ind in np.dtype(self.sim_type).names)
            except BaseException:
                raise ValueError(f"{sim_ind[0]} not a name in given sim_type")
            self.sim_name = sim_ind
            self.goal = jnp.ones(1)
        elif isinstance(sim_ind, tuple):
            try:
                assert sim_ind[0] in np.dtype(self.sim_type).names
            except BaseException:
                raise ValueError(f"{sim_ind[0]} not a name in given sim_type")
            try:
                assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
            except BaseException:
                raise ValueError(f"{sim_ind[1]} not an index of {sim_ind[0]}")
            self.sim_name = sim_ind[0]
            sim_size = self.sim_type[sim_ind[0]].shape[0]
            self.goal = jnp.eye(sim_size)[sim_ind[1]]
        else:
            raise TypeError("sim_ind must be a str or (str, int) tuple")
        # Check for optional input
        self.bound = bound
        if bound_type.strip().lower() not in ('lower', 'upper'):
            raise ValueError("bound type must be 'upper' or 'lower'")
        if bound_type.strip().lower() == 'lower':
            self.goal *= -1.0
            self.bound *= -1.0
        return

    def __call__(self, x, sx):
        """ Define constraint evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The (negative when feasible) slack in this constraint
            for the input sx (x is unused).

        """

        # These lines look silly, but do not change them.
        # They are needed to gracefully cast all possible shapes to scalar.
        fx = 0.0
        fx += jnp.dot(sx[self.sim_name], self.goal)
        return fx - self.bound


class SumOfSimSquaresBound(CompositeFunction):
    """ Class for constraining the sum-of-squared simulation outputs.

    If upper bounding:

    ```
    ConstraintFunction(x, sx):
        return sum([sx[i]**2 for all i]) - upper_bound
    ```

    If lower bounding:

    ```
    def ConstraintFunction(x, sx):
        return lower_bound - sum([sx[i]**2 for all i])
    ```

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_inds, bound_type='upper', bound=0)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` evaluate the slack (negative values are feasible).

    """

    def __init__(self, des_type, sim_type, sim_inds,
                 bound_type='upper', bound=0.0):
        """ Constructor for SumOfSimSquaresBound class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_inds (list): The list of names (or name, index pairs for sims
                with more than one output) of the simulation outputs to sum
                over.

            bound_type (str): Either 'lower' to lower-bound or 'upper' to
                upper-bound. Defaults to 'upper'.

            bound (float): The lower/upper bound for this constraint.
                Defaults to 0.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if not isinstance(sim_inds, list):
            raise TypeError("sim_inds must be a list of ints, tuples, or " +
                            "strings")
        self.sim_names = []
        self.sim_inds = []
        for sim_ind in sim_inds:
            if isinstance(sim_ind, str):
                try:
                    assert (sim_ind in np.dtype(self.sim_type).names)
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                if sim_ind in self.sim_names:
                    self.sim_inds[self.sim_inds.index(sim_ind)] += 1.0
                else:
                    self.sim_names.append(sim_ind)
                    self.sim_inds.append(jnp.ones(1))
            elif isinstance(sim_ind, tuple):
                try:
                    assert sim_ind[0] in np.dtype(self.sim_type).names
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                try:
                    assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
                except BaseException:
                    raise ValueError(f"{sim_ind[1]} not an index of "
                                     f"{sim_ind[0]}")
                sim_size = self.sim_type[sim_ind[0]].shape[0]
                if sim_ind[0] in self.sim_names:
                    self.sim_inds[self.sim_names.index(sim_ind[0])] += \
                            jnp.eye(sim_size)[sim_ind[1]]
                else:
                    self.sim_names.append(sim_ind[0])
                    self.sim_inds.append(jnp.eye(sim_size)[sim_ind[1]])
            else:
                raise TypeError("Each sim_ind must be a str or str,int tuple")
        # Check for optional input
        self.bound = bound
        if bound_type.lower() not in ('lower', 'upper'):
            raise ValueError("bound type must be 'upper' or 'lower'")
        if bound_type.lower() == 'lower':
            self.goal = -1.0
            self.bound *= -1.0
        else:
            self.goal = 1.0
        return

    def __call__(self, x, sx):
        """ Define constraint evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The (negative when feasible) slack in this constraint
            for the input sx (x is unused).

        """

        fx = 0.0
        for sn, si in zip(self.sim_names, self.sim_inds):
            fx += jnp.dot(sx[sn] ** 2, si)
        return fx * self.goal - self.bound


class SumOfSimsBound(CompositeFunction):
    """ Class for bounding the (absolute) sum of simulation outputs.

    If upper bounding:

    ```
    def const_func(x, sx):
        return sum([sx[i] for all i]) - upper_bound
    ```

    If lower bounding:

    ```
    def const_func(x, sx):
        return lower_bound - sum([sx[i] for all i])
    ```

    If upper bounding absolute sum:

    ```
    def const_func(x, sx):
        return sum([abs(sx[i]) for all i]) - upper_bound
    ```

    If lower bounding absolute sum:

    ```
    def const_func(x, sx):
        return lower_bound - sum([abs(sx[i]) for all i])
    ```

    Also supports derivative usage.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, sim_inds,
                  bound_type='upper', bound=0, absolute=False)``
     * ``__call__(x, sim)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` evaluate the slack (negative values are feasible).

    """

    def __init__(self, des_type, sim_type, sim_inds,
                 bound_type='upper', bound=0.0, absolute=False):
        """ Constructor for SumOfSimsBound class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            sim_inds (list): The list of names (or name, index pairs for sims
                with more than one output) of the simulation outputs to sum
                over.

            bound_type (str): Either 'lower' to lower-bound or 'upper' to
                upper-bound. Defaults to 'upper'.

            bound (float): The lower/upper bound for this constraint.
                Defaults to 0.

            absolute (bool): True to bound absolute sum, False to
                bound raw sum. Defaults to False.

        """

        super().__init__(des_type, sim_type)
        # Check additional inputs
        if not isinstance(sim_inds, list):
            raise TypeError("sim_inds must be a list of ints, tuples, or " +
                            "strings")
        self.sim_names = []
        self.sim_inds = []
        for sim_ind in sim_inds:
            if isinstance(sim_ind, str):
                try:
                    assert (sim_ind in np.dtype(self.sim_type).names)
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                if sim_ind in self.sim_names:
                    self.sim_inds[self.sim_inds.index(sim_ind)] += 1.0
                else:
                    self.sim_names.append(sim_ind)
                    self.sim_inds.append(jnp.ones(1))
            elif isinstance(sim_ind, tuple):
                try:
                    assert sim_ind[0] in np.dtype(self.sim_type).names
                except BaseException:
                    raise ValueError(f"{sim_ind[0]} not in given sim_type")
                try:
                    assert 0 <= sim_ind[1] < self.sim_type[sim_ind[0]].shape[0]
                except BaseException:
                    raise ValueError(f"{sim_ind[1]} not an index of "
                                     f"{sim_ind[0]}")
                sim_size = self.sim_type[sim_ind[0]].shape[0]
                if sim_ind[0] in self.sim_names:
                    self.sim_inds[self.sim_names.index(sim_ind[0])] += \
                            jnp.eye(sim_size)[sim_ind[1]]
                else:
                    self.sim_names.append(sim_ind[0])
                    self.sim_inds.append(jnp.eye(sim_size)[sim_ind[1]])
            else:
                raise TypeError("Each sim_ind must be a str or str,int tuple")
        # Check for optional input
        self.bound = bound
        if bound_type.lower() not in ('lower', 'upper'):
            raise ValueError("bound type must be 'upper' or 'lower'")
        if bound_type.lower() == 'lower':
            self.goal = -1.0
            self.bound *= -1.0
        else:
            self.goal = 1.0

        def id_func(x): return np.ones(x.size)

        def abs_func(x): return jnp.sign(x)

        if not isinstance(absolute, bool):
            raise TypeError("absolute must be a bool type, not " +
                            str(type(absolute)))
        if absolute:
            self.absolute = abs_func
        else:
            self.absolute = id_func
        return

    def __call__(self, x, sx):
        """ Define constraint evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The (negative when feasible) slack in this constraint
            for the input sx (x is unused).

        """

        fx = 0.0
        for sn, si in zip(self.sim_names, self.sim_inds):
            fx += jnp.sum(self.absolute(sx[sn]) * sx[sn] * si)
        return fx * self.goal - self.bound


class SingleSimBoundGradient(SingleSimBound):
    """ Gradient class for SingleSimBound.

    Inherits from the ``SingleSimBound`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient output of the simulation at self.sim_ind
            given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        ds[self.sim_name] = self.goal
        return dx, ds


class SumOfSimSquaresBoundGradient(SumOfSimSquaresBound):
    """ Gradient class for SumOfSimSquaresBound.

    Inherits from the ``SumOfSimSquaresBound`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of the sum of squared simulation outputs
            given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        for sn, si in zip(self.sim_names, self.sim_inds):
            ds[sn] = sx[sn] * si * 2.0 * self.goal
        return dx, ds


class SumOfSimsBoundGradient(SumOfSimsBound):
    """ Gradient class for SumOfSimsBound.

    Inherits from the ``SumOfSimsBound`` class, but overwrites the
    ``__call__`` method to return the gradients wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ Define gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of the sum of (absolute) simulation
            outputs given the input (x, sx) wrt x and sx, respectively.

        """

        dx, ds = {}, {}
        for name in self.des_type.names:
            dx[name] = jnp.zeros(1)
        for name in self.sim_type.names:
            size = max(sum(self.sim_type[name].shape), 1)
            ds[name] = jnp.zeros(size)
        for sn, si in zip(self.sim_names, self.sim_inds):
            ds[sn] = self.absolute(sx[sn]) * si * self.goal
        return dx, ds



""" This module contains utilities (helper functions) that are used throughout
the viz tool.

"""

import pandas as pd
import logging


def export_file(fig, plot_name, file_type):
    """ Export image of figure to working directory.

        Args:
            fig (plotly.graph_objects.Figure): The figure to export.

            plot_name (string): Set the filename of the image file.

            file_type (string): Set the image file type.
             - 'html' - Export as .html file.
             - 'pdf' - Export as .pdf file.
             - 'svg' - Export as .svg file.
             - 'eps' - Export as .eps file
               if the poppler dependency is installed.
             - 'jpeg' - Export as .jpeg file.
             - 'png' - Export as .png file.
             - 'webp' - Export as .webp file.

    """

    if file_type == 'html':
        fig.write_html(plot_name + ".html")
        logging.info("exported graph as .html")
    elif file_type == 'pdf':
        fig.write_image(plot_name + ".pdf")
        logging.info("exported graph as .pdf")
    elif file_type == 'svg':
        fig.write_image(plot_name + ".svg")
        logging.info("exported graph as .svg")
    elif file_type == 'eps':
        fig.write_image(plot_name + ".eps")
        logging.info("exported graph as .eps")
    elif file_type == 'jpeg':
        fig.write_image(plot_name + ".jpeg")
        logging.info("exported graph as .jpeg")
    elif file_type == 'png':
        fig.write_image(plot_name + ".png")
        logging.info("exported graph as .png")
    elif file_type == 'webp':
        fig.write_image(plot_name + ".webp")
        logging.info("exported graph as .webp")


def set_plot_name(db):
    """ Provide a default graph title.

        Args:
            db (string): Graph contents inform title.
             - 'pf' - Set plot name to "Pareto Front"
             - 'obj' - Set plot name to "Objective Data"

        Returns:
            plot_name (string): The default plot name.

    """

    if db == 'pf':
        plot_name = "Pareto Front"
    elif db == 'obj':
        plot_name = "Objective Data"
    return plot_name


def set_database(moop, db, points):
    """ Choose which points from MOOP object to plot.

        Args:
            db (string): Set dataset.
             - 'pf' - Set Pareto Front as dataset.
             - 'obj' - Set objective data as dataset.

            points (string): Filter traces from dataset by constraint score.
             - 'constraint_satisfying' - Include only points that
               satisfy every constraint.
             - 'constraint_violating' - Include only points that
               violate any constraint.
             - 'all' - Include all points in dataset.
             - 'none' - Include no points in dataset.

        Returns:
            df (Pandas dataframe): A 2D dataframe containing post-filter
            data from the MOOP.

    """

    if db == 'pf':
        database = pd.DataFrame(moop.getPF())
    elif db == 'obj':
        database = pd.DataFrame(moop.getObjectiveData())
    if moop.getConstraintType() is None:
        df = database
    else:
        if points == 'constraint_satisfying':
            constraints = moop.getConstraintType().names
            df = database.copy(deep=True)
            for constraint in constraints:
                indices = df[df[constraint] > 0].index
                df.drop(indices, inplace=True)
                df.reset_index(inplace=True, drop=True)
        elif points == 'constraint_violating':
            constraints = moop.getConstraintType().names
            df = database.copy(deep=True)
            for constraint in constraints:
                indices = df[df[constraint] <= 0].index
                df.drop(indices, inplace=True)
                df.reset_index(inplace=True, drop=True)
        elif points == 'all':
            df = database
        elif points == 'none':
            df = database[0:0]
    return df


def set_hover_info(database, i):
    """ Customize information in hover label for trace i.

        Args:
            database (Pandas dataframe): A 2D dataframe containing the
                traces to be graphed.

            i (int): An index indicating the row where the trace
                we're labeling is located.

        Returns:
            hover_info (string): An HTML-format string to display when
            users hover over trace i.

    """

    hover_info = ""
    for key in database.columns:
        hover_info += str(key)
        hover_info += ": "
        hover_info += str(database[key][i])
        hover_info += "<br>"
        # since plotly is JavaScript-based
        # it uses HTML string formatting
    return hover_info


def check_inputs(db,
                 output,
                 points,
                 height,
                 width,
                 font,
                 fontsize,
                 background_color,
                 screenshot,
                 image_export_format,
                 data_export_format,
                 dev_mode,
                 pop_up,
                 port,
                 ):
    """ Check keyword inputs to user-facing functions for validity

        Args:
            db: The item passed to the 'db' keyword in a user-facing function.
                If db cannot be cast to a string valued 'pf' or 'obj',
                a ValueError is raised.

            output: The item passed to the 'output' keyword in a
                user-facing function.
                If output cannot be cast to a string corresponding to one of
                the supported output filetypes, a ValueError is raised.

            points: The item passed to the 'points' keyword in a
                user-facing function.
                If points cannot be cast to a string corresponding to one of
                the supported constraint filters, a ValueError is raised.

            height: The item passed to the 'height' keyword in a user-facing
                function.
                If height is not the default string 'auto' or cannot be cast
                to an int
                of value greater than one, a ValueError is raised.

            width: The item passed to the 'width' keyword in a user-facing
                function.
                If width is not the default string 'auto' or cannot be cast
                to an int
                of value greater than one, a ValueError is raised.

            font: The item passed to the 'font' keyword in a user-facing
                function.
                If font cannot be cast to a string, a ValueError is raised.

            fontsize: The item passed to the 'fontsize' keyword in a
                user-facing function.
                If fontsize is not the default value 'auto' or cannot be cast
                to an int
                of value between 1 and 100 inclusive, a ValueError is raised.

            background_color: The item passed to the 'background_color'
                keyword in a user-facing function.
                If background_color cannot be cast to a string, a ValueError
                is raised.

            screenshot: The item passed to the 'screenshot' keyword in a
                user-facing function.
                If screenshot cannot be cast to a string corresponding to one
                of the supported
                screenshot filetypes, a ValueError is raised.

            image_export_format: The item passed to the 'image_export_format'
                keyword in a user-facing function.
                If image_export_format cannot be cast to a string
                corresponding to one of the supported
                image_export_format filetypes, a ValueError is raised.

            data_export_format: The item passed to the 'data_export_format'
                keyword in a user-facing function.
                If data_export_format cannot be cast to a string corresponding
                to one of the supported
                data_export_format filetypes, a ValueError is raised.

            data_export_format: The item passed to the 'data_export_format'
                keyword in a user-facing function.
                If data_export_format cannot be cast to a string corresponding
                to one of the supported
                data_export_format filetypes, a ValueError is raised.

            dev_mode: The item passed to the 'dev_mode' keyword in a
                user-facing function.
                If dev_mode cannot be cast to one of the Boolean values True
                and False, a ValueError is raised.

            pop_up: The item passed to the 'pop_up' keyword in a user-facing
                function.
                If pop_up cannot be cast to one of the Boolean values True and
                False, a ValueError is raised.

            port: The item passed to the 'port' keyword in a user-facing
                function.
                If port cannot be cast to a string beginning with 'http', a
                ValueError is raised.

        Raises:
            A ValueError if any of the values passed by a user to a keyword in
            a user-facing function are judged invalid.

    """

    try:
        if (str(db) == 'pf' or str(db) == 'obj'):
            pass
        else:
            raise ValueError(str(db) + " is an invalid value for 'db'")
    except BaseException:
        raise ValueError(str(db) + " is an invalid value for 'db'")

    try:
        if (str(output) == 'dash' or
                str(output) == 'html' or
                str(output) == 'svg' or
                str(output) == 'pdf' or
                str(output) == 'eps' or
                str(output) == 'jpeg' or
                str(output) == 'png' or
                str(output) == 'webp'):
            pass
        else:
            raise ValueError(str(output) + " is an invalid value for 'output'")
    except BaseException:
        raise ValueError(str(output) + " is an invalid value for 'output'")

    try:
        if (str(points) == 'constraint_satisfying' or
                str(points) == 'constraint_violating' or
                str(points) == 'all' or
                str(points) == 'none'):
            pass
        else:
            raise ValueError(str(points) + " is an invalid value for 'points'")
    except BaseException:
        raise ValueError(str(points) + " is an invalid value for 'points'")

    if (height == 'auto' or int(height) >= 1):
        pass
    else:
        raise ValueError(str(height) + " is an invalid value for 'height'")

    try:
        if (width == 'auto' or int(width) >= 1):
            pass
        else:
            raise ValueError(str(width) + " is an invalid value for 'width'")
    except BaseException:
        raise ValueError(str(width) + " is an invalid value for 'width'")

    try:
        if (str(type(str(font))) == "<class 'str'>"):
            pass
        else:
            raise ValueError(str(font) + " is an invalid value for 'font'")
    except BaseException:
        raise ValueError(str(font) + " is an invalid value for 'font'")

    try:
        if (fontsize == 'auto' or
           (int(fontsize) >= 1 and int(fontsize) <= 100)):
            pass
        else:
            message = str(fontsize)
            message += " is an invalid value for 'fontsize'"
            raise ValueError(message)
    except BaseException:
        message = str(fontsize)
        message += " is an invalid value for 'fontsize'"
        raise ValueError(message)

    try:
        if (str(type(str(background_color))) == "<class 'str'>"):
            pass
        else:
            message = str(background_color)
            message += " is an invalid value for 'background_color'"
            raise ValueError(message)
    except BaseException:
        message = str(background_color)
        message += " is an invalid value for 'background_color'"
        raise ValueError(message)

    try:
        if (screenshot == 'png' or
                screenshot == 'svg' or
                screenshot == 'jpeg' or
                screenshot == 'webp'):
            pass
        else:
            message = str(screenshot)
            message += " is an invalid value for 'screenshot'"
            raise ValueError(message)
    except BaseException:
        message = str(screenshot)
        message += " is an invalid value for 'screenshot'"
        raise ValueError(message)

    try:
        if (image_export_format == 'html' or
                image_export_format == 'svg' or
                image_export_format == 'pdf' or
                image_export_format == 'eps' or
                image_export_format == 'jpeg' or
                image_export_format == 'png' or
                image_export_format == 'webp'):
            pass
        else:
            message = str(image_export_format)
            message += " is an invalid value for 'image_export_format'"
            raise ValueError(message)
    except BaseException:
        message = str(image_export_format)
        message += " is an invalid value for 'image_export_format'"
        raise ValueError(message)

    try:
        if (data_export_format == 'json' or
                data_export_format == 'csv'):
            pass
        else:
            message = str(data_export_format)
            message += " is an invalid value for 'data_export_format'"
            raise ValueError(message)
    except BaseException:
        message = str(data_export_format)
        message += " is an invalid value for 'data_export_format'"
        raise ValueError(message)

    try:
        if (dev_mode or not dev_mode):
            pass
        else:
            message = str(dev_mode)
            message += " is an invalid value for 'dev_mode'"
            raise ValueError(message)
    except BaseException:
        message = str(dev_mode)
        message += " is an invalid value for 'dev_mode'"
        raise ValueError(message)

    try:
        if (pop_up or not pop_up):
            pass
        else:
            raise ValueError(str(pop_up) + " is an invalid value for 'pop_up'")
    except BaseException:
        raise ValueError(str(pop_up) + " is an invalid value for 'pop_up'")

    try:
        if (port[0:4] == 'http'):
            pass
        else:
            raise ValueError(str(port) + " is an invalid value for 'port'")
    except BaseException:
        raise ValueError(str(port) + " is an invalid value for 'port'")



def test_SingleSimBound():
    """ Test the SingleSimBound() function.

    Initialize an objective then evaluate.

    """

    from parmoo.constraints import SingleSimBound, SingleSimBoundGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 2)]
    # Create input vectors
    x = np.zeros(1, dtype=xtype)[0]
    sx = np.zeros(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][1] = 3.0
    # Create the objective and gradient functions
    obj_func1 = SingleSimBound(xtype, stype, 'sim1', bound_type='lower')
    grad_func1 = SingleSimBoundGradient(xtype, stype, 'sim1',
                                        bound_type='lower')
    obj_func2 = SingleSimBound(xtype, stype, ('sim2', 1), bound_type='upper')
    grad_func2 = SingleSimBoundGradient(xtype, stype, ('sim2', 1),
                                        bound_type='upper')
    # Test function evaluations
    assert (obj_func1(x, sx) == -2.0)
    assert (obj_func2(x, sx) == 3.0)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3"]
    dx, ds = grad_func1(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -1.0)
    assert (np.all(ds["sim2"] == 0.0))
    dx, ds = grad_func2(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == 0.0)
    assert (np.all(ds["sim2"] == np.eye(2)[1]))


def test_SumOfSimSquaresBound():
    """ Test the SumOfSimSquaresBound() function.

    Initialize an objective then evaluate.

    """

    from parmoo.constraints import SumOfSimSquaresBound
    from parmoo.constraints import SumOfSimSquaresBoundGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8"), ("x4", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 4)]
    # Create input vectors
    x = np.ones(1, dtype=xtype)[0]
    sx = np.ones(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][:] = 2.0
    # Create the objective and gradient functions
    sim_list = ['sim1', ('sim2', 0), ('sim2', 1), ('sim2', 2)]
    obj_func = SumOfSimSquaresBound(xtype, stype, sim_list, bound_type="lower")
    grad_func = SumOfSimSquaresBoundGradient(xtype, stype, sim_list,
                                             bound_type="lower")
    # Test function evaluation
    assert (np.abs(obj_func(x, sx) + 16.0) < 1.0e-8)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3", "x4"]
    dx, ds = grad_func(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -4.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4)) * 4.0))


def test_SumOfSimsBound():
    """ Test the SumOfSimsBound() function.

    Initialize an objective then evaluate.

    """

    from parmoo.constraints import SumOfSimsBound, SumOfSimsBoundGradient
    import numpy as np

    # Create named dtypes
    xtype = [("x1", "f8"), ("x2", "f8"), ("x3", "f8"), ("x4", "f8")]
    stype = [("sim1", "f8"), ("sim2", "f8", 4)]
    # Create input vectors
    x = np.ones(1, dtype=xtype)[0]
    sx = np.ones(1, dtype=stype)[0]
    sx['sim1'] = 2.0
    sx['sim2'][:] = -2.0
    # Create the objective functions
    sim_list = ['sim1', ('sim2', 0), ('sim2', 1), ('sim2', 2)]
    obj_func1 = SumOfSimsBound(xtype, stype, sim_list, bound_type="lower")
    grad_func1 = SumOfSimsBoundGradient(xtype, stype, sim_list,
                                        bound_type="lower")
    obj_func2 = SumOfSimsBound(xtype, stype, sim_list,
                               bound_type='upper', absolute=True)
    grad_func2 = SumOfSimsBoundGradient(xtype, stype, sim_list,
                                        bound_type='upper', absolute=True)
    # Test function evaluation
    assert (np.abs(obj_func1(x, sx) - 4.0) < 1.0e-8)
    assert (np.abs(obj_func2(x, sx) - 8.0) < 1.0e-8)
    # Test dx and ds evaluation
    xkeys = ["x1", "x2", "x3", "x4"]
    dx, ds = grad_func1(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == -1.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4))))
    dx, ds = grad_func2(x, sx)
    assert (all([dx[i] == 0.0 for i in xkeys]))
    assert (ds["sim1"] == 1.0)
    assert (np.all(ds["sim2"] == (np.eye(4)[3] - np.ones(4))))


if __name__ == "__main__":
    test_SingleSimBound()
    test_SumOfSimSquaresBound()
    test_SumOfSimsBound()



""" This module contains objective function implementations of the DTLZ test
suite, as described in:

Deb, Thiele, Laumanns, and Zitzler. "Scalable test problems for
evolutionary multiobjective optimization" in Evolutionary Multiobjective
Optimization, Theoretical Advances and Applications, Ch. 6 (pp. 105--145).
Springer-Verlag, London, UK, 2005. Abraham, Jain, and Goldberg (Eds).

Since DTLZ[1-7] depended upon kernel functions (implemented in
parmoo.simulations.dtlz), each of these problems is implemented here
as an algebraic, differentiable objective, with the kernel function output
as an input. The problems DTLZ8 and DTLZ9 do not support this modification,
so they are omitted.

TODO: DTLZ5, DTLZ6, and DTLZ7 have not yet been added.

To use this module, first import one or more of the following simulation/kernel
functions from parmoo.simulations.dtlz:
 * ``g1_sim``
 * ``g2_sim``
 * ``g3_sim``
 * ``g4_sim``

The 7 DTLZ problems included here are:
 * ``dtlz1_obj``
 * ``dtlz2_obj``
 * ``dtlz3_obj``
 * ``dtlz4_obj``

"""

from jax import numpy as jnp
import numpy as np
from parmoo.structs import CompositeFunction
from parmoo.util import to_array, from_array


class dtlz1_obj(CompositeFunction):
    """ Class defining the DTLZ1 objectives.

    Use this class in combination with the g1_sim() class from the
    parmoo.simulations.dtlz module

    DTLZ1 has a linear Pareto front, with all nondominated points
    on the hyperplane F_1 + F_2 + ... + F_o = 0.5.
    DTLZ1 has 11^k - 1 "local" Pareto fronts where k = n - m + 1, and
    1 "global" Pareto front.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, obj_ind)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` method performs an evaluation of the DTLZ1 problem.

    """

    def __init__(self, des_type, sim_type, obj_ind, num_obj=3):
        """ Constructor for DTLZ1 class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            obj_ind (int): The index of the DTLZ1 objective to return.

            num_obj (int, optional): The number of objectives for this problem.
                Note that this effects the calculation of the objective value,
                but still only a single objective output is created per
                instance of this class. To add all objectives, create
                num_obj instances with obj_ind = 0, ..., num_obj - 1.

        """

        super().__init__(des_type, sim_type)
        if self.m != 1:
            raise ValueError("DTLZ1 only supports 1 simulation output, " +
                             "but " + str(self.m) + " were given")
        if not isinstance(obj_ind, int):
            raise TypeError("optional input obj_ind must have the int type")
        if obj_ind < 0:
            raise ValueError("obj_ind cannot be negative")
        self.obj_ind = obj_ind
        if not isinstance(num_obj, int):
            raise TypeError("optional input num_obj must have the int type")
        if num_obj < 0:
            raise ValueError("num_obj cannot be negative")
        self.o = num_obj
        return

    def __call__(self, x, sx):
        """ DTLZ1 objective function evaluation at the specified index.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of this objective for the input x.

        """

        # Roll x into xx and sx into ssx
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate
        i = self.obj_ind
        fx = (1 + ssx) * jnp.prod(xx[:self.o - 1 - i]) / 2
        if i > 0:
            fx *= (1 - xx[self.o - 1 - i])
        return fx


class dtlz2_obj(CompositeFunction):
    """ Class defining the DTLZ2 objectives.

    Use this class in combination with the g2_sim() class from the
    parmoo.simulations.dtlz module.

    DTLZ2 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ2 has no "local" Pareto fronts, besides the true Pareto front.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, obj_ind)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` method performs an evaluation of the DTLZ2 problem.

    """

    def __init__(self, des_type, sim_type, obj_ind, num_obj=3):
        """ Constructor for DTLZ2 class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            obj_ind (int): The index of the DTLZ2 objective to return.

            num_obj (int, optional): The number of objectives for this problem.
                Note that this effects the calculation of the objective value,
                but still only a single objective output is created per
                instance of this class. To add all objectives, create
                num_obj instances with obj_ind = 0, ..., num_obj - 1.

        """

        super().__init__(des_type, sim_type)
        if self.m != 1:
            raise ValueError("DTLZ2 only supports 1 simulation output, " +
                             "but " + str(self.m) + " were given")
        if not isinstance(obj_ind, int):
            raise TypeError("optional input obj_ind must have the int type")
        if obj_ind < 0:
            raise ValueError("obj_ind cannot be negative")
        self.obj_ind = obj_ind
        if not isinstance(num_obj, int):
            raise TypeError("optional input num_obj must have the int type")
        if num_obj < 0:
            raise ValueError("num_obj cannot be negative")
        self.o = num_obj
        return

    def __call__(self, x, sx):
        """ DTLZ2 objective function evaluation at the specified index.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of this objective for the input x.

        """

        # Extract x into xx and sx into ssx
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate
        i = self.obj_ind
        fx = (1 + ssx) * jnp.prod(jnp.cos(xx[:self.o - i - 1] * jnp.pi / 2))
        if i > 0:
            fx *= jnp.sin(jnp.pi * xx[self.o - i - 1] / 2)
        return fx


class dtlz3_obj(CompositeFunction):
    """ Class defining the DTLZ3 objectives.

    Use this class in combination with the g1_sim() class from the
    parmoo.simulations.dtlz module.

    DTLZ3 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ3 has 3^k - 1 "local" Pareto fronts where k = n - o + 1, and
    1 "global" Pareto front.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, obj_ind)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` method performs an evaluation of the DTLZ3 problem.

    """

    def __init__(self, des_type, sim_type, obj_ind, num_obj=3):
        """ Constructor for DTLZ3 class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            obj_ind (int): The index of the DTLZ3 objective to return.

            num_obj (int, optional): The number of objectives for this problem.
                Note that this effects the calculation of the objective value,
                but still only a single objective output is created per
                instance of this class. To add all objectives, create
                num_obj instances with obj_ind = 0, ..., num_obj - 1.

        """

        super().__init__(des_type, sim_type)
        if self.m != 1:
            raise ValueError("DTLZ3 only supports 1 simulation output, " +
                             "but " + str(self.m) + " were given")
        if not isinstance(obj_ind, int):
            raise TypeError("optional input obj_ind must have the int type")
        if obj_ind < 0:
            raise ValueError("obj_ind cannot be negative")
        self.obj_ind = obj_ind
        if not isinstance(num_obj, int):
            raise TypeError("optional input num_obj must have the int type")
        if num_obj < 0:
            raise ValueError("num_obj cannot be negative")
        self.o = num_obj
        return

    def __call__(self, x, sx):
        """ DTLZ3 objective function evaluation at the specified index.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of this objective for the input x.

        """

        # Extract x into xx and sx into ssx
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate
        i = self.obj_ind
        fx = (1 + ssx) * jnp.prod(jnp.cos(xx[:self.o - i - 1] * jnp.pi / 2))
        if i > 0:
            fx *= jnp.sin(jnp.pi * xx[self.o - i - 1] / 2)
        return fx


class dtlz4_obj(CompositeFunction):
    """ Class defining the DTLZ4 objectives.

    Use this class in combination with the g2_sim() class from the
    parmoo.simulations.dtlz module.

    DTLZ4 has a concave Pareto front, given by the unit sphere in
    objective space, restricted to the positive orthant.
    DTLZ4 has no "local" Pareto fronts, besides the true Pareto front,
    but by tuning the optional parameter alpha, one can adjust the
    solution density, making it harder for MOO algorithms to produce
    a uniform distribution of solutions.

    Contains 2 methods:
     * ``__init__(des_type, sim_type, obj_ind)``
     * ``__call__(x, sx)``

    The ``__init__`` method inherits from the CompositeFunction ABC.

    The ``__call__`` method performs an evaluation of the DTLZ4 problem.

    """

    def __init__(self, des_type, sim_type, obj_ind, num_obj=3, alpha=100.0):
        """ Constructor for DTLZ4 class.

        Args:
            des_type (np.dtype): The numpy.dtype of the design variables.

            sim_type (np.dtype): The numpy.dtype of the simulation outputs.

            obj_ind (int): The index of the DTLZ4 objective to return.

            num_obj (int, optional): The number of objectives for this problem.
                Note that this effects the calculation of the objective value,
                but still only a single objective output is created per
                instance of this class. To add all objectives, create
                num_obj instances with obj_ind = 0, ..., num_obj - 1.

            alpha (optional, float or int): The uniformity parameter used for
                controlling the uniformity of the distribution of solutions
                across the Pareto front. Must be greater than or equal to 1.
                A value of 1 results in DTLZ2. Default value is 100.0.

        """

        super().__init__(des_type, sim_type)
        if self.m != 1:
            raise ValueError("DTLZ4 only supports 1 simulation output, " +
                             "but " + str(self.m) + " were given")
        if not isinstance(obj_ind, int):
            raise TypeError("optional input obj_ind must have the int type")
        if obj_ind < 0:
            raise ValueError("obj_ind cannot be negative")
        self.obj_ind = obj_ind
        if not isinstance(num_obj, int):
            raise TypeError("optional input num_obj must have the int type")
        if num_obj < 0:
            raise ValueError("num_obj cannot be negative")
        self.o = num_obj
        if not (isinstance(alpha, int) or isinstance(alpha, float)):
            raise TypeError("alpha must be a numeric type")
        if alpha < 1:
            raise ValueError("alpha must be at least 1")
        self.alpha = alpha
        return

    def __call__(self, x, sx):
        """ DTLZ4 objective function evaluation at the specified index.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            float: The output of this objective for the input x.

        """

        # Extract x into xx and sx into ssx
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate
        i = self.obj_ind
        fx = (1 + ssx) * np.prod(np.cos(xx[:self.o - i - 1] ** self.alpha *
                                 np.pi / 2))
        if i > 0:
            fx *= np.sin(xx[self.o - i - 1] ** self.alpha * np.pi / 2)
        return fx


class dtlz1_grad(dtlz1_obj):
    """ Class defining the DTLZ1 gradient.

    Inherits from ``dtlz1_obj``, but overwrites the ``__call__`` method to
    evaluate the gradient wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ DTLZ1 gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of this objective for the input (x, sx)
            wrt x and sx, respectively.

        """

        # Roll x into xx and sx into ssx
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate grad wrt x
        i = self.obj_ind
        dx = jnp.zeros(self.n)
        for j in range(self.o - i - 1):
            dxj = (jnp.prod(xx[0:j]) * jnp.prod(xx[j + 1:self.o - i - 1])
                   * (1 + ssx) / 2)
            if i > 0:
                dxj *= (1.0 - xx[self.o - i - 1])
            dx = dx.at[j].set(dxj)
        if i != 0:
            j = self.o - i - 1
            dxj = -(jnp.prod(xx[:self.o - i - 1]) * (1 + ssx) / 2)
            dx = dx.at[j].set(dxj)
        # Evaluate grad wrt sx
        ds = jnp.prod(xx[:self.o - i - 1]) / 2
        if i > 0:
            ds *= (1 - xx[self.o - i - 1])
        return from_array(dx, self.des_type), from_array(ds, self.sim_type)


class dtlz2_grad(dtlz2_obj):
    """ Class defining the DTLZ2 gradient.

    Inherits from ``dtlz2_obj``, but overwrites the ``__call__`` method to
    evaluate the gradient wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ DTLZ2 gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of this objective for the input (x, sx)
            wrt x and sx, respectively.

        """

        # Extract x into xx and sx into ssx, if names are used
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate derivative wrt xx
        i = self.obj_ind
        dx = jnp.zeros(self.n)
        for j in range(self.o - i - 1):
            dxj = (jnp.prod(jnp.cos(xx[:j] * jnp.pi / 2)) *
                   (-jnp.pi / 2) * jnp.sin(xx[j] * jnp.pi / 2) *
                   jnp.prod(jnp.cos(xx[j + 1:self.o - i - 1] * jnp.pi / 2)) *
                   (1 + ssx))
            if i > 0:
                dxj *= jnp.sin(xx[self.o - i - 1] * jnp.pi / 2)
            dx = dx.at[j].set(dxj)
        if i > 0:
            j = self.o - i - 1
            dxj = (jnp.prod(jnp.cos(xx[:self.o-i-1] * jnp.pi / 2)) *
                   (1 + ssx) * (jnp.pi / 2) *
                   jnp.cos(xx[self.o - i - 1] * jnp.pi / 2))
            dx = dx.at[j].set(dxj)
        # Evaluate derivative wrt ssx
        ds = jnp.prod(jnp.cos(xx[:self.o - i - 1] * jnp.pi / 2))
        if i > 0:
            ds *= jnp.sin(jnp.pi * xx[self.o - i - 1] / 2)
        return from_array(dx, self.des_type), from_array(ds, self.sim_type)


class dtlz3_grad(dtlz3_obj):
    """ Class defining the DTLZ3 gradient.

    Inherits from ``dtlz3_obj``, but overwrites the ``__call__`` method to
    evaluate the gradient wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ DTLZ3 gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of this objective for the input (x, sx)
            wrt x and sx, respectively.

        """

        # Extract x into xx and sx into ssx, if names are used
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate derivative wrt xx
        i = self.obj_ind
        dx = jnp.zeros(self.n)
        for j in range(self.o - i - 1):
            dxj = (jnp.prod(jnp.cos(xx[:j] * jnp.pi / 2)) *
                   (-jnp.pi / 2) * jnp.sin(xx[j] * jnp.pi / 2) *
                   jnp.prod(jnp.cos(xx[j + 1:self.o - i - 1] * jnp.pi / 2)) *
                   (1 + ssx))
            if i > 0:
                dxj *= jnp.sin(xx[self.o - i - 1] * jnp.pi / 2)
            dx = dx.at[j].set(dxj)
        if i > 0:
            j = self.o - i - 1
            dxj = (jnp.prod(jnp.cos(xx[:self.o-i-1] * jnp.pi / 2)) *
                   (1 + ssx) * (jnp.pi / 2) *
                   jnp.cos(xx[self.o - i - 1] * jnp.pi / 2))
            dx = dx.at[j].set(dxj)
        # Evaluate derivative wrt ssx
        ds = jnp.prod(jnp.cos(xx[:self.o - i - 1] * jnp.pi / 2))
        if i > 0:
            ds *= jnp.sin(jnp.pi * xx[self.o - i - 1] / 2)
        return from_array(dx, self.des_type), from_array(ds, self.sim_type)


class dtlz4_grad(dtlz4_obj):
    """ Class defining the DTLZ4 gradient.

    Inherits from ``dtlz4_obj``, but overwrites the ``__call__`` method to
    evaluate the gradient wrt x and sx, respectively.

    """

    def __call__(self, x, sx):
        """ DTLZ4 gradient evaluation.

        Args:
            x (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of a design point to evaluate.

            sx (dict or structured array): A Python dictionary or numpy
                structured array containing the names (as keys) and
                corresponding values of the simulation output(s) at x.

        Returns:
            dict, dict: The gradient of this objective for the input (x, sx)
            wrt x and sx, respectively.

        """

        # Extract x into xx and sx into ssx, if names are used
        xx = to_array(x, self.des_type)
        ssx = to_array(sx, self.sim_type)[0]
        # Evaluate derivative wrt xx
        i = self.obj_ind
        dx = jnp.zeros(self.n)
        for j in range(self.o - i - 1):
            dxj = (jnp.prod(jnp.cos(xx[:j] ** self.alpha * jnp.pi / 2))
                   * (-jnp.pi * self.alpha / 2)
                   * xx[j] ** (self.alpha - 1)
                   * jnp.sin(xx[j] ** self.alpha * jnp.pi / 2) *
                   jnp.prod(jnp.cos(xx[j+1:self.o-i-1] ** self.alpha *
                            jnp.pi / 2)) * (1 + ssx))
            if i > 0:
                dxj *= jnp.sin(xx[self.o - i - 1] ** self.alpha * jnp.pi / 2)
            dx = dx.at[j].set(dxj)
        if i > 0:
            j = self.o - i - 1
            dxj = (jnp.prod(jnp.cos(xx[0:self.o-i-1] ** self.alpha *
                            jnp.pi / 2)) * (1 + ssx) *
                   (jnp.pi * self.alpha / 2) *
                   xx[j] ** (self.alpha - 1) *
                   jnp.cos(xx[j] ** self.alpha * jnp.pi / 2))
            dx = dx.at[j].set(dxj)
        # Evaluate derivative wrt ssx
        ds = jnp.prod(jnp.cos(xx[:self.o - i - 1] ** self.alpha *
                              jnp.pi / 2))
        if i > 0:
            ds *= jnp.sin(xx[self.o - i - 1] ** self.alpha * jnp.pi / 2)
        return from_array(dx, self.des_type), from_array(ds, self.sim_type)




""" Implementations of Pattern Search (PS) and its variations.

This module contains implementations of the SurrogateOptimizer ABC, which
are based on pattern search.

Note that these strategies are all gradient-free, and therefore do not
require any gradients to be defined. This makes them friendly for
getting started.

The classes include:
 * ``LocalSurrogate_PS`` -- A multistart pattern search (PS) algorithm
 * ``GlobalSurrogate_PS`` -- global random search, followed by LocalPS

"""

import jax
from jax import numpy as jnp
import numpy as np
from parmoo.structs import SurrogateOptimizer
from parmoo.util import xerror
from scipy.stats.qmc import LatinHypercube


class LocalSurrogate_PS(SurrogateOptimizer):
    """ Use multistart Pattern Search to solve surrogate problem locally.

    Applies PS to the surrogate problem, in order to identify design
    points that are locally Pareto optimal, with respect to the surrogate
    problem. Sorts poll directions by most recently used and attempts to
    step in promising directions in late iterations.

    """

    # Slots for the LocalSurrogate_PS class
    __slots__ = ['n', 'lb', 'ub', 'acquisitions', 'budget', 'constraints',
                 'objectives', 'simulations', 'setTR',
                 'penalty_func', 'sim_sd', 'restarts', 'momentum', 'q_ind',
                 'prev_centers', 'des_tols', 'targets', 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the LocalSurrogate_PS class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure. It may contain the following:
                 * opt_budget (int): The PS iteration limit (default: 1000).
                 * opt_restarts (int): Number of multistart restarts per
                   scalarization (default: n+1).

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.n = lb.size
        self.lb = lb
        self.ub = ub
        self.q_ind = 0
        # Check that the contents of hyperparams are legal
        if 'opt_restarts' in hyperparams:
            if isinstance(hyperparams['opt_restarts'], int):
                if hyperparams['opt_restarts'] < 1:
                    raise ValueError("hyperparams['opt_restarts'] "
                                     "must be positive")
                else:
                    self.restarts = hyperparams['opt_restarts']
            else:
                raise TypeError("hyperparams['opt_restarts'] "
                                "must be an integer")
        else:
            self.restarts = self.n + 1
        # Check that the contents of hyperparams are legal
        if 'opt_budget' in hyperparams:
            if isinstance(hyperparams['opt_budget'], int):
                if hyperparams['opt_budget'] < 1:
                    raise ValueError("hyperparams['opt_budget'] "
                                     "must be positive")
                else:
                    self.budget = hyperparams['opt_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.budget = 1000
        # Check that the contents of hyperparams are legal
        if 'opt_momentum' in hyperparams:
            if isinstance(hyperparams['opt_momentum'], float):
                if 0 <= hyperparams['opt_momentum'] < 1:
                    self.momentum = hyperparams['opt_momentum']
                else:
                    raise ValueError("hyperparams['opt_momentum'] "
                                     "must be in [0, 1)")
            else:
                raise TypeError("hyperparams['opt_momentum'] "
                                "must be a float")
        else:
            self.momentum = 9e-1
        self.eps = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        if 'des_tols' in hyperparams:
            if isinstance(hyperparams['des_tols'], np.ndarray):
                if hyperparams['des_tols'].size != self.n:
                    raise ValueError("the length of hyperparpams['des_tols']"
                                     " must match the length of lb and ub")
                if not np.all(hyperparams['des_tols']):
                    raise ValueError("all entries in hyperparams['des_tols']"
                                     " must be greater than 0")
            else:
                raise TypeError("hyperparams['des_tols'] must be an array.")
            self.des_tols = np.asarray(hyperparams['des_tols'])
        else:
            self.des_tols = (np.ones(self.n) * self.eps)
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.acquisitions = []
        self.prev_centers = []
        self.targets = []
        return

    def _obj_func(self, x_in):
        """ A wrapper for the objective function and acquisition.

        Args:
            x_in (np.ndarray): A 1d array with the design point to evaluate.

        Returns:
            float: The result of acquisition.scalarize(f(x_in, sim(x_in))).

        """

        sx_in = self.simulations(x_in)
        if self.acquisitions[self.q_ind].useSD():
            sdx_in = self.sim_sd(x_in)
        else:
            sdx_in = jnp.zeros(sx_in.size)
        fx_in = self.penalty_func(x_in, sx_in)
        ax = self.acquisitions[self.q_ind].scalarize(fx_in, x_in,
                                                     sx_in, sdx_in)
        return ax

    def _checkTR(self, center):
        """ Check the recommended trust region for a new center. """

        # Search the history for the given radius
        rad = np.zeros(self.n)
        for (ci, ri) in reversed(self.prev_centers):
            if np.all(np.abs(center - np.asarray(ci)) < self.des_tols):
                rad[:] = np.asarray(ri)
                break
        # If not found in the history initialize
        if np.all(rad == 0):
            rad = (self.ub - self.lb) * 0.1
            rad = np.maximum(rad, self.des_tols)
        return rad

    def _checkTargets(self):
        """ Use internal list of targets to check and update the TR radii """

        for ti in self.targets:
            # Decay all "missed" targets' TR radii
            ci, ri, _, _ = ti
            ri = np.maximum(ri * 0.5, self.des_tols)
            # Update the TR history
            found = False
            j = 0
            for (cj, rj) in reversed(self.prev_centers):
                j -= 1
                if np.all(np.abs(cj - ci) < self.des_tols):
                    self.prev_centers[j][-1] = ri
            if not found:
                self.prev_centers.append([ci, ri])
        # Reset the list of targets for next iteration
        self.targets = []
        return

    def returnResults(self, x, fx, sx, sdx):
        """ Collect the results of a function evaluation.

        Args:
            x (np.ndarray): The design point evaluated.

            fx (np.ndarray): The objective function values at x.

            sx (np.ndarray): The simulation function values at x.

            sdx (np.ndarray): The standard deviation in the simulation
                outputs at x.

        """

        for i, ti in enumerate(self.targets):
            j = ti[3]
            fxj = self.acquisitions[j].scalarize(fx, x, sx, sdx)
            # Remove any targets that have been "hit"
            if fxj < ti[2]:
                del self.targets[i]
        return

    def solve(self, x):
        """ Solve the surrogate problem in a trust region via pattern search.

        Args:
            x (np.ndarray): A 2d array containing a list of feasible
                design points used to warm start the search.

        Returns:
            np.ndarray: A 2d numpy.ndarray of potentially efficient design
            points that were found by the PS optimizer.

        """

        # Check that x is legal
        if self.n != x.shape[1]:
            raise ValueError("The columns of x must match n")
        elif len(self.acquisitions) != x.shape[0]:
            raise ValueError("The rows of x must match the number " +
                             "of acquisition functions")
        # Initialize an empty list of results
        result = []
        lb_tmp = np.zeros(self.n)
        ub_tmp = np.ones(self.n)
        self._checkTargets()
        # For each acquisition function
        for j, acquisition in enumerate(self.acquisitions):
            # Create a new trust region
            rad = self._checkTR(x[j, :])
            self.setTR(x[j, :], rad)
            lb_tmp[:] = np.maximum(self.lb[:], x[j, :] - rad)
            ub_tmp[:] = np.minimum(self.ub[:], x[j, :] + rad)
            # Recompile the objective function
            self.q_ind = j
            mesh_tol = max(self.eps,
                           np.min((ub_tmp - lb_tmp) * np.sqrt(self.eps)))
            try:
                obj_func = jax.jit(self._obj_func)
                _ = obj_func(x[j])
            except BaseException:
                obj_func = self._obj_func
            # Get a candidate
            xj, fj = _accelerated_pattern_search(self.n, lb_tmp,
                                                 ub_tmp, x[j],
                                                 obj_func,
                                                 ibudget=self.budget,
                                                 mesh_tol=mesh_tol,
                                                 momentum=self.momentum,
                                                 istarts=self.restarts,
                                                 np_rng=self.np_rng)
            result.append(xj)
            # We need to remember this "target" for later
            self.targets.append([x[j, :], rad, fj, j])
        self.objectives = None
        self.constraints = None
        self.penalty_func = None
        return np.asarray(result)

    def save(self, filename):
        """ Save important data from this class so that it can be reloaded.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data should be saved.

        """

        import json

        # Serialize PS object in dictionary
        ps_state = {'n': self.n,
                    'budget': self.budget,
                    'restarts': self.restarts,
                    'momentum': self.momentum,
                    'q_ind': self.q_ind}
        # Serialize numpy.ndarray objects
        ps_state['lb'] = self.lb.tolist()
        ps_state['ub'] = self.ub.tolist()
        ps_state['des_tols'] = self.des_tols.tolist()
        # Flatten arrays
        ps_state['prev_centers'] = []
        for (ci, ri) in self.prev_centers:
            ps_state['rev_centers'].append([ci.tolist(), ri.tolist()])
        ps_state['targets'] = []
        for ti in self.targets:
            ps_state['targets'].append([ti[0].tolist(), ti[1].tolist(),
                                        ti[2], ti[3]])
        # Save file
        with open(filename, 'w') as fp:
            json.dump(ps_state, fp)
        return

    def load(self, filename):
        """ Reload important data into this class after a previous save.

        Args:
            filename (string): The relative or absolute path to the file
                where all reload data has been saved.

        """

        import json

        # Load file
        with open(filename, 'r') as fp:
            ps_state = json.load(fp)
        # Deserialize PS object from dictionary
        self.n = ps_state['n']
        self.budget = ps_state['budget']
        self.restarts = ps_state['restarts']
        self.momentum = ps_state['momentum']
        self.q_ind = ps_state['q_ind']
        # Deserialize numpy.ndarray objects
        self.lb = np.array(ps_state['lb'])
        self.ub = np.array(ps_state['ub'])
        self.des_tols = np.array(ps_state['des_tols'])
        # Extract history arrays
        self.prev_centers = []
        for (ci, ri) in ps_state['prev_centers']:
            self.prev_centers.append([np.array(ci), np.array(ri)])
        self.targets = []
        for ti in ps_state['targets']:
            self.targets.append([np.array(ti[0]), np.array(ti[1]),
                                 ti[2], ti[3]])
        return


class GlobalSurrogate_PS(SurrogateOptimizer):
    """ Use randomized search globally followed by a local PS.

    Use ``RandomSearch`` to globally search the design space followed
    ``LocalSurrogate_PS`` to refine the potentially efficient solutions.

    """

    # Slots for the GlobalSurrogate_PS class
    __slots__ = ['n', 'o', 'lb', 'ub', 'acquisitions', 'constraints',
                 'objectives', 'simulations', 'setTR',
                 'penalty_func', 'opt_budget', 'gps_budget', 'sim_sd',
                 'momentum', 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the GlobalPS class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for the
                optimization procedure. It may contain the following:
                 * opt_budget (int): The function evaluation budget
                   (default: 1500)
                 * ps_budget (int): The number of the total opt_budget
                   evaluations that will be used by PS (default: 2/3
                   of opt_budget).

        Returns:
            SurrogateOptimizer: A new SurrogateOptimizer object.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.n = lb.size
        self.o = o
        self.lb = lb
        self.ub = ub
        # Check that the contents of hyperparams are legal
        if 'opt_budget' in hyperparams:
            if isinstance(hyperparams['opt_budget'], int):
                if hyperparams['opt_budget'] < 1:
                    raise ValueError("hyperparams['opt_budget'] "
                                     "must be positive")
                else:
                    self.opt_budget = hyperparams['opt_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.opt_budget = 1500
        # Check that the contents of hyperparams are legal
        self.eps = np.sqrt(jnp.finfo(jnp.ones(1)).eps)
        if 'gps_budget' in hyperparams:
            if isinstance(hyperparams['gps_budget'], int):
                if hyperparams['gps_budget'] < 1 or \
                   hyperparams['gps_budget'] >= self.opt_budget:
                    raise ValueError("hyperparams['gps_budget'] "
                                     "must be between 1 and "
                                     "hyperparams['opt_budget']")
                else:
                    self.gps_budget = hyperparams['gps_budget']
            else:
                raise TypeError("hyperparams['opt_budget'] "
                                "must be an integer")
        else:
            self.gps_budget = int(2 * self.opt_budget / 3)
        # Check that the contents of hyperparams are legal
        if 'opt_momentum' in hyperparams:
            if isinstance(hyperparams['opt_momentum'], float):
                if 0 <= hyperparams['opt_momentum'] < 1:
                    self.momentum = hyperparams['opt_momentum']
                else:
                    raise ValueError("hyperparams['opt_momentum'] "
                                     "must be in [0, 1)")
            else:
                raise TypeError("hyperparams['opt_momentum'] "
                                "must be a float")
        else:
            self.momentum = 9e-1
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.acquisitions = []
        return

    def _obj_func(self, x_in):
        """ A wrapper for the objective function and acquisition.

        Args:
            x_in (np.ndarray): A 1d array with the design point to evaluate.

        Returns:
            float: The result of acquisition.scalarize(f(x_in, sim(x_in))).

        """

        sx_in = np.asarray(self.simulations(x_in))
        if self.acquisitions[self.q_ind].useSD():
            sdx_in = np.asarray(self.sim_sd(x_in))
        else:
            sdx_in = np.zeros(sx_in.size)
        fx_in = np.asarray(self.penalty_func(x_in, sx_in)).flatten()
        ax = self.acquisitions[self.q_ind].scalarize(fx_in, x_in,
                                                     sx_in, sdx_in)
        return ax

    def solve(self, x):
        """ Solve the surrogate problem by using random search followed by PS.

        Args:
            x (np.ndarray): A 2d array containing a list of feasible
                design points used to warm start the search.

        Returns:
            np.ndarray: A 2d numpy.ndarray containing a list of potentially
            efficient design points that were found by the optimizers.

        """

        from parmoo.util import updatePF

        # Check that x is legal
        if self.n != x.shape[1]:
            raise ValueError("The columns of x must match n")
        elif len(self.acquisitions) != x.shape[0]:
            raise ValueError("The rows of x must match the number " +
                             "of acquisition functions")
        # Create an infinite trust region
        rad = np.ones(self.n) * np.inf
        self.setTR(self.lb, rad)
        # Perform a random search globally
        batch_size = 1000
        data = {'x_vals': np.zeros((batch_size, self.n)),
                'f_vals': np.zeros((batch_size, self.o)),
                'c_vals': np.zeros((batch_size, 0))}
        # Loop over batch size until k == budget
        k = 0
        nondom = {}
        search_budget = self.opt_budget = self.gps_budget
        while (k < search_budget):
            # Check how many new points to generate
            k_new = min(search_budget, k + batch_size) - k
            if k_new < batch_size:
                data['x_vals'] = np.zeros((k_new, self.n))
                data['f_vals'] = np.zeros((k_new, self.o))
                data['c_vals'] = np.zeros((k_new, 0))
            # Randomly generate k_new new points
            for i in range(k_new):
                if i < len(self.acquisitions):
                    xi = x[i, :]
                else:
                    xi = (self.np_rng.random(self.n) *
                          (self.ub[:] - self.lb[:]) + self.lb[:])
                sxi = self.simulations(xi)
                data['x_vals'][i, :] = xi[:]
                data['f_vals'][i, :] = self.penalty_func(xi, sxi)
            # Update the PF
            nondom = updatePF(data, nondom)
            k += k_new
        f_vals = []
        # Loop over acquisition functions and perform pattern search
        result = []
        for j, acq in enumerate(self.acquisitions):
            if acq.useSD():
                f_vals = [acq.scalarize(fi, xi, self.simulations(xi),
                                        self.sim_sd(xi))
                          for fi, xi in zip(nondom['f_vals'],
                                            nondom['x_vals'])]
            else:
                m = self.simulations(nondom['x_vals'][0]).size
                f_vals = [acq.scalarize(fi, xi, self.simulations(xi),
                                        np.zeros(m))
                          for fi, xi in zip(nondom['f_vals'],
                                            nondom['x_vals'])]
            imin = np.argmin(np.asarray([f_vals]))
            x0 = nondom['x_vals'][imin, :].copy()
            # Recompile the objective function
            self.q_ind = j
            mesh_tol = max(self.eps,
                           np.min((self.ub - self.lb) * np.sqrt(self.eps)))
            try:
                obj_func = jax.jit(self._obj_func)
                _ = obj_func(x[j])
            except BaseException:
                obj_func = self._obj_func
            # Get a candidate
            xj, fj = _accelerated_pattern_search(self.n, self.lb,
                                                 self.ub, x0,
                                                 obj_func,
                                                 ibudget=self.gps_budget,
                                                 mesh_start=0.1,
                                                 mesh_tol=mesh_tol,
                                                 momentum=self.momentum,
                                                 istarts=1)
            result.append(xj)
        self.objectives = None
        self.constraints = None
        self.penalty_func = None
        return np.asarray(result)


def _accelerated_pattern_search(n, lb, ub, x0, obj_func, ibudget,
                                mesh_start=None, mesh_tol=1.0e-8,
                                momentum=0.9, istarts=1, np_rng=None):
    """ Solve the optimization problem min obj_func(x) over x in [lb, ub].

    Uses pattern search with an additional search direction inspired by
    Nesterov's momentum.

    Args:
        n (int): The dimension of the search space.

        lb (np.ndarray): A 1D array of length n specifying lower bounds on
            the search space.

        ub (np.ndarray): A 1D array of length n specifying upper bounds on
            the search space.

        x0 (np.ndarray): A 1D array of length n specifying the initial
            location to sample when warm-starting search.

        obj_func (function): A function that takes a 1D numpy.ndarray of
            size n as input (x) and returns the value of f(x).

        ibudget (int): The iteration limit for pattern search. The total
            number of calls to obj_func could be up to 2n * ibudget + 1,
            but typically will be much less.

        mesh_start (float or numpy.ndarray, optional): The initial mesh
            spacing. If an array is given, must be 1D of size n. Defaults
            to (ub - lb) / (istarts - 1).

        mesh_tol (float or numpy.ndarray, optional): The tolerance for the
            mesh. If an array is given, must be 1D of size n. Defaults to
            1.0e-8.

        momentum (float, optional): The decay rate on the momentum term in
            the accelerated poll direction. Defaults to 0.9.

        istarts (int, optional): Number of times to (re)start if
            using the multistart option. Defaults to 1 (no restarts).

        np_rng (numpy.random.Generator, optional): An instance of a numpy
            generator object that will be used for generating consistent
            restarts. Defaults to None, which results in a new generator
            being created.

    Returns:
        numpy.ndarray(n): The best x observed so far (minimizer of obj_func).

    """

    # Initialize O(n)-sized work arrays and constants
    f_min = np.zeros(istarts)
    m_tmp = np.zeros(n)
    mesh_size = np.zeros(n)
    x_center = np.zeros(n)
    x_min = np.zeros((istarts, n))
    x_tmp = np.zeros(n)
    mesh_start_array = np.zeros(n)
    if mesh_start is None:
        mesh_start_array[:] = 1 / istarts
    elif isinstance(mesh_start_array, float):
        mesh_start_array[:] = mesh_start * (ub[:] - lb[:])
    else:
        mesh_start_array[:] = mesh_start
    if istarts > 1:
        lhs = LatinHypercube(n, seed=np_rng).random(istarts - 1)
    if np_rng is None:
        np_rng = np.random.default_rng()
    # Check working tolerance (unit roundoff)
    mu = jnp.finfo(jnp.ones(1)).eps
    root_mu = np.sqrt(mu)
    # Loop over all starts
    for kk in range(istarts):
        # Reset the mesh dimensions
        mesh_size[:] = mesh_start_array[:]
        mesh = np.vstack((np.zeros((1, n)),
                          np.diag(ub[:] - lb[:]),
                          -np.diag(ub[:] - lb[:])))
        # Evaluate the starting point or random point
        if kk == 0:
            x_min[kk, :] = x0[:]
        else:
            x_min[kk, :] = lhs[kk - 1] * (ub - lb) + lb
        f0 = obj_func(x_min[kk])
        f_tol = max(min(abs(f0)**2, root_mu), mu)
        f_min[kk] = f0
        # Take n+1 iterations to get "momentum" started
        k_start = 0
        for k in range(ibudget):
            improve = False
            x_center[:] = x_min[kk, :]
            for i, mi in enumerate(mesh[1:, :]):
                # Evaluate x + mi
                x_tmp[:] = x_center[:] + mi[:] * mesh_size[:]
                if _check_bounds(x_tmp, lb, ub):
                    f_tmp = np.inf
                else:
                    f_tmp = obj_func(x_tmp)
                # Check for improvement
                if _check_improv(f_min[kk], f_tmp, f_tol):
                    f_min[kk] = f_tmp
                    x_min[kk, :] = x_tmp[:]
                    m_min = i + 1
                    improve = True
            # If there was improvement, shuffle the directions
            if improve:
                iswitch = np.where(np.abs(mesh[m_min, :]) > 1.0e-8)[0]
                mesh[0, :] *= momentum
                mesh[0, iswitch] = mesh[m_min, iswitch]
                m_tmp[:] = mesh[m_min, :]
                mesh[2:m_min+1, :] = mesh[1:m_min, :]
                mesh[1, :] = m_tmp[:]
                # Update k_start and break the "warm-up" loop
                k_start += 1
                if k_start >= n+1:
                    break
            # If no improvement, decay the mesh down to the tolerance
            else:
                if np.any(mesh_size[:] < mesh_tol):
                    break
                else:
                    mesh_size[:] *= 0.5
        # Take the remaining iterations
        for k in range(k_start, ibudget):
            improve = False
            x_center[:] = x_min[kk, :]
            for i, mi in enumerate(mesh[:, :]):
                # Evaluate x + mi
                x_tmp[:] = x_center[:] + np.rint(mi[:]) * mesh_size[:]
                if _check_bounds(x_tmp, lb, ub):
                    f_tmp = np.inf
                else:
                    f_tmp = obj_func(x_tmp)
                # Check for improvement
                if _check_improv(f_min[kk], f_tmp, f_tol):
                    f_min[kk] = f_tmp
                    x_min[kk, :] = x_tmp[:]
                    m_min = i
                    improve = True
                    break
            # If there was improvement, update moment and shuffle directions
            if improve:
                if m_min > 0:
                    iswitch = np.where(np.abs(mesh[m_min, :]) > 1.0e-8)[0]
                    mesh[0, :] *= momentum
                    mesh[0, iswitch] = mesh[m_min, iswitch]
                    m_tmp[:] = mesh[m_min, :]
                    mesh[2:m_min+1, :] = mesh[1:m_min, :]
                    mesh[1, :] = m_tmp[:]
            # If no improvement, decay the mesh down to the tolerance
            else:
                if np.any(mesh_size[:] < mesh_tol):
                    break
                else:
                    mesh_size[:] *= 0.5
    imin = np.argmin(f_min)
    return x_min[imin].copy(), f_min[imin]


@jax.jit
def _check_improv(f_min, f_tmp, f_tol):
    return f_min - f_tmp > f_tol


@jax.jit
def _check_bounds(x_tmp, lb, ub):
    return np.any((x_tmp > ub) + (x_tmp < lb))




""" Implementations of the epsilon-constraint-style scalarizations.

This module contains implementations of the AcquisitionFunction ABC, which
use the epsilon constraint method.

The classes include:
 * ``RandomConstraint`` (randomly set a ub for all but 1 objective)

"""

import inspect
from jax import numpy as jnp
import numpy as np
from parmoo.structs import AcquisitionFunction
from parmoo.util import xerror
from scipy import stats, integrate


class RandomConstraint(AcquisitionFunction):
    """ Improve upon a randomly set target point.

    Randomly sets a target point inside the current Pareto front.
    Attempts to improve one of the objective values by reformulating
    all other objectives as constraints, upper bounded by their target
    value.

    """

    # Slots for the RandomConstraint class
    __slots__ = ['n', 'o', 'lb', 'ub', 'f_ub', 'weights', 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the RandomConstraint class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for tuning
                the acquisition function.

        Returns:
            RandomConstraint: A new RandomConstraint scalarizer.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.o = o
        # Set the design variable count
        self.n = np.size(lb)
        # Initialize the objective/design bounds
        self.f_ub = np.zeros(self.o)
        self.weights = np.zeros(self.o)
        self.ub = ub
        self.lb = lb
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.eps = jnp.finfo(jnp.ones(1)).eps
        return

    def useSD(self):
        """ Query whether this method uses uncertainties.

        When False, allows users to shortcut expensive uncertainty
        computations.

        """

        return False

    def setTarget(self, data, penalty_func):
        """ Randomly generate a target based on current nondominated points.

        Args:
            data (dict): A dictionary specifying the current function
                evaluation database. It contains two mandatory fields:
                 * 'x_vals' (numpy.ndarray): A 2d array containing the
                   list of design points.
                 * 'f_vals' (numpy.ndarray): A 2d array containing the
                   corresponding list of objective values.

            penalty_func (function): A function of one (x) or two (x, sx)
                inputs that evaluates the (penalized) objectives.

        Returns:
            numpy.ndarray: A 1d array containing the 'best' feasible starting
            point for the scalarized problem (if any previous evaluations
            were feasible) or the point in the existing database that is
            most nearly feasible.

        """

        from parmoo.util import updatePF

        # Check whether any data was given
        no_data = False
        # Check for illegal input from data
        if not isinstance(data, dict):
            raise TypeError("data must be a dict")
        else:
            if ('x_vals' in data) != ('f_vals' in data):
                raise AttributeError("if x_vals is a key in data, then " +
                                     "f_vals must also appear")
            elif 'x_vals' in data:
                if data['x_vals'] is not None and data['f_vals'] is not None:
                    if data['x_vals'].shape[0] != data['f_vals'].shape[0]:
                        raise ValueError("x_vals and f_vals must be equal " +
                                         "length")
                    if data['x_vals'].shape[1] != self.n:
                        raise ValueError("The rows of x_vals must have " +
                                         "length n")
                    if data['f_vals'].shape[1] != self.o:
                        raise ValueError("The rows of f_vals must have " +
                                         "length o")
                else:
                    no_data = True
            else:
                no_data = True
        # Check whether penalty_func() has an appropriate signature
        if callable(penalty_func):
            if len(inspect.signature(penalty_func).parameters) not in [1, 2]:
                raise ValueError("penalty_func() must accept 1 or 2"
                                 + " inputs")
        else:
            raise TypeError("penalty_func() must be callable")
        if no_data:
            # If data is empty, then the Pareto front is empty
            pf = {'x_vals': np.zeros((0, self.n)),
                  'f_vals': np.zeros((0, self.o)),
                  'c_vals': np.zeros((0, 1))}
        else:
            # Get the Pareto front
            pf = updatePF(data, {})
        # If data is empty, randomly select weights and starting point
        if no_data:
            self.weights = -np.log(1.0 - np.random.random_sample(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            self.f_ub[:] = np.inf
            # Randomly select a starting point
            x_start = (self.np_rng.random(self.n) * (self.ub - self.lb)
                       + self.lb)
            return x_start
        # If data is nonempty but pf is empty, use a penalty to select
        elif pf is None or pf['x_vals'].shape[0] == 0:
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            self.f_ub[:] = np.inf
            # Check for "most feasible" starting x
            x_best = np.zeros(data['x_vals'].shape[1])
            p_best = np.inf
            for xi, fi, ci in zip(data['x_vals'], data['f_vals'],
                                  data['c_vals']):
                p_temp = np.sum(fi) / np.sqrt(self.eps) + np.sum(ci)
                if p_temp < p_best:
                    x_best = xi
                    p_best = p_temp
            return x_best
        else:
            # Randomly select pts in the convex hull of the nondominated pts
            ipts = self.np_rng.integers(0, pf['f_vals'].shape[0], size=self.o)
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            target = np.dot(self.weights, pf['f_vals'][ipts, :])
            fi = pf['f_vals'][ipts[0], :]
            # Set the bounds
            self.f_ub[:] = np.inf
            self.weights[:] = self.eps ** 0.25
            for j in range(self.o):
                # If fi[j] is less than target[j], this is a bound
                if fi[j] + 0.00000001 < target[j]:
                    self.f_ub[j] = target[j]
                # Otherwise, it is an objective
                else:
                    self.weights[j] = 1.0
            # Normalize the weights
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            # The corresponding x_val is feasible by construction
            return pf['x_vals'][ipts[0], :]

    def scalarize(self, f_vals, x_vals, s_vals_mean, s_vals_sd):
        """ Scalarize a vector of function values using the current bounds.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

            x_vals (np.ndarray): A 1D array specifying a vector the design
                point corresponding to f_vals (unused by this method).

            s_vals_mean (np.ndarray): A 1D array specifying the expected
                simulation outputs for the x value being scalarized
                (unused by this method).

            s_vals_sd (np.ndarray): A 1D array specifying the standard
                deviation for each of the simulation outputs (unused by
                this method).

        Returns:
            float: The scalarized value.

        """

        result = jnp.dot(f_vals, self.weights)
        for i in range(self.o):
            result += 10.0 * jnp.maximum(f_vals[i] - self.f_ub[i], 0)
        return result


class EI_RandomConstraint(AcquisitionFunction):
    """ Expected improvement of a randomly set target point.

    Randomly sets a target point inside the current Pareto front.
    Attempts to improve one of the objective values by reformulating
    all other objectives as constraints, upper bounded by their target
    value. Uses surrogate uncertainties to maximize expected improvement
    in the target objective subject to constraints.

    """

    # Slots for the RandomConstraint class
    __slots__ = ['n', 'o', 'lb', 'ub', 'f_ub', 'weights', 'best', 'f',
                 'np_rng', 'eps']

    def __init__(self, o, lb, ub, hyperparams):
        """ Constructor for the RandomConstraint class.

        Args:
            o (int): The number of objectives.

            lb (numpy.ndarray): A 1d array of lower bounds for the design
                region. The number of design variables is inferred from the
                dimension of lb.

            ub (numpy.ndarray): A 1d array of upper bounds for the design
                region. The dimension must match ub.

            hyperparams (dict): A dictionary of hyperparameters for tuning
                the acquisition function. Including
                 * mc_sample_size (int): The number of samples to use for
                   Monte Carlo integration (defaults to 10 * m ** 2).

        Returns:
            RandomConstraint: A new RandomConstraint scalarizer.

        """

        # Check inputs
        xerror(o=o, lb=lb, ub=ub, hyperparams=hyperparams)
        self.o = o
        # Set the design variable count
        self.n = np.size(lb)
        # Initialize the objective/design bounds
        self.f_ub = np.zeros(self.o)
        self.weights = np.zeros(self.o)
        self.ub = ub
        self.lb = lb
        # Check the hyperparameter dictionary for custom MC sample size
        if 'mc_sample_size' in hyperparams.keys():
            self.sample_size = hyperparams['mc_sample_size']
        else:
            self.sample_size = None
        # Check the hyperparameter dictionary for random generator
        if 'np_random_gen' in hyperparams:
            if isinstance(hyperparams['np_random_gen'], np.random.Generator):
                self.np_rng = hyperparams['np_random_gen']
            else:
                raise TypeError("When present, hyperparams['np_random_gen'] "
                                "must be an instance of the class "
                                "numpy.random.Generator")
        else:
            self.np_rng = np.random.default_rng()
        self.eps = jnp.finfo(jnp.ones(1)).eps
        return

    def useSD(self):
        """ Query whether this method uses uncertainties.

        When False, allows users to shortcut expensive uncertainty
        computations.

        """

        return True

    def setTarget(self, data, penalty_func):
        """ Randomly generate a target based on current nondominated points.

        Args:
            data (dict): A dictionary specifying the current function
                evaluation database. It contains two mandatory fields:
                 * 'x_vals' (numpy.ndarray): A 2d array containing the
                   list of design points.
                 * 'f_vals' (numpy.ndarray): A 2d array containing the
                   corresponding list of objective values.

            penalty_func (function): A function of one (x) or two (x, sx)
                inputs that evaluates the (penalized) objectives.

        Returns:
            numpy.ndarray: A 1d array containing the 'best' feasible starting
            point for the scalarized problem (if any previous evaluations
            were feasible) or the point in the existing database that is
            most nearly feasible.

        """

        from parmoo.util import updatePF

        # Check whether any data was given
        no_data = False
        # Check for illegal input from data
        if not isinstance(data, dict):
            raise TypeError("data must be a dict")
        else:
            if ('x_vals' in data) != ('f_vals' in data):
                raise AttributeError("if x_vals is a key in data, then " +
                                     "f_vals must also appear")
            elif 'x_vals' in data:
                if data['x_vals'] is not None and data['f_vals'] is not None:
                    if data['x_vals'].shape[0] != data['f_vals'].shape[0]:
                        raise ValueError("x_vals and f_vals must be equal " +
                                         "length")
                    if data['x_vals'].shape[1] != self.n:
                        raise ValueError("The rows of x_vals must have " +
                                         "length n")
                    if data['f_vals'].shape[1] != self.o:
                        raise ValueError("The rows of f_vals must have " +
                                         "length o")
                else:
                    no_data = True
            else:
                no_data = True
        # Check whether penalty_func() has an appropriate signature
        if callable(penalty_func):
            if len(inspect.signature(penalty_func).parameters) not in [1, 2]:
                raise ValueError("penalty_func() must accept 1 or 2"
                                 + " inputs")
        else:
            raise TypeError("penalty_func() must be callable")
        # Save the penalty function for later
        self.f = penalty_func
        if no_data:
            # If data is empty, then the Pareto front is empty
            pf = {'x_vals': np.zeros((0, self.n)),
                  'f_vals': np.zeros((0, self.o)),
                  'c_vals': np.zeros((0, 1))}
        else:
            # Get the Pareto front
            pf = updatePF(data, {})
        # If data is empty, randomly select weights and starting point
        if no_data:
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            self.f_ub[:] = np.inf
            # Randomly select a starting point
            x_start = (self.np_rng.random(self.n) * (self.ub - self.lb)
                       + self.lb)
            return x_start
        # If data is nonempty but pf is empty, use a penalty to select
        elif pf is None or pf['x_vals'].shape[0] == 0:
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            self.f_ub[:] = np.inf
            # Check for "most feasible" starting x
            x_best = np.zeros(data['x_vals'].shape[1])
            p_best = np.inf
            for xi, fi, ci in zip(data['x_vals'], data['f_vals'],
                                  data['c_vals']):
                p_temp = np.sum(fi) / np.sqrt(self.eps) + np.sum(ci)
                if p_temp < p_best:
                    x_best = xi
                    p_best = p_temp
            return x_best
        else:
            # Randomly select pts in the convex hull of the nondominated pts
            ipts = self.np_rng.integers(0, pf['f_vals'].shape[0], size=self.o)
            self.weights = -np.log(1.0 - self.np_rng.random(self.o))
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            target = np.dot(self.weights, pf['f_vals'][ipts, :])
            fi = pf['f_vals'][ipts[0], :]
            # Set the bounds
            self.f_ub[:] = np.inf
            self.weights[:] = 0.0001
            for j in range(self.o):
                # If fi[j] is less than target[j], this is a bound
                if fi[j] + 0.00000001 < target[j]:
                    self.f_ub[j] = target[j]
                # Otherwise, it is an objective
                else:
                    self.weights[j] = 1.0
            # Normalize the weights
            self.weights[:] = self.weights[:] / np.linalg.norm(self.weights)
            # The corresponding x_val is feasible by construction
            self.best = np.dot(pf['f_vals'][ipts[0], :], self.weights)
            return pf['x_vals'][ipts[0], :]

    def scalarize(self, f_vals, x_vals, s_vals_mean, s_vals_sd):
        """ Scalarize a vector of function values using the current bounds.

        Args:
            f_vals (numpy.ndarray): A 1d array specifying the function
                values to be scalarized.

            x_vals (np.ndarray): A 1D array specifying a vector the design
                point corresponding to f_vals (unused by this method).

            s_vals_mean (np.ndarray): A 1D array specifying the expected
                simulation outputs for the x value being scalarized
                (unused by this method).

            s_vals_sd (np.ndarray): A 1D array specifying the standard
                deviation for each of the simulation outputs (unused by
                this method).

        Returns:
            float: The scalarized value.

        """

        # If the feasible set was empty, just use the given fi
        if self.best is None:
            result = np.dot(f_vals, self.weights)
            for i in range(self.o):
                if f_vals[i] > self.f_ub[i]:
                    result = result + 10.0 * (f_vals[i] - self.f_ub[i])
            return result
        # If the feasible set was nonempty and the number of sim outs is 1,
        # then calculate the EI over the best seen feasible fi for the
        # current penalty function
        elif s_vals_mean.size == 1:
            # Construct the distribution for sampling
            s_cov = stats.Covariance.from_diagonal(s_vals_sd**2)
            s_dist = stats.multivariate_normal(mean=s_vals_mean, cov=s_cov)

            def weighted_f(sx):
                """ Calculates the pdf-weighted value of f at sx """

                fx = self.f(x_vals, np.array([sx]))
                # Add penalty
                for j in range(self.o):
                    if fx[j] > self.f_ub[j]:
                        fx[:] = fx[:] + 10.0 * (fx[j] - self.f_ub[j])
                result = min(np.dot(fx, self.weights) - self.best, 0.0)
                return result * s_dist.pdf(np.array([sx]))

            y = integrate.quad(weighted_f, -np.inf, np.inf)
            return y[0]
        # If there is at least one feasible point and the number of sim outs
        # is greater than 1, then evaluate the EI over fi* via MC integration
        else:
            if self.sample_size is None:
                self.sample_size = int(10 * s_vals_mean.size ** 2)
            # Construct the distribution for sampling
            s_cov = stats.Covariance.from_diagonal(s_vals_sd)
            s_dist = stats.multivariate_normal(mean=s_vals_mean, cov=s_cov)
            result = 0.0
            # Loop over sample size
            for i in range(self.sample_size):
                s_vals = s_dist.rvs().flatten()
                fi = self.f(x_vals, s_vals)
                # Add penalty
                for j in range(self.o):
                    if fi[j] > self.f_ub[j]:
                        fi[:] = fi[:] + 10.0 * (fi[j] - self.f_ub[j])
                result = min(np.dot(fi, self.weights) - self.best, 0.0)
            result /= self.sample_size
            return result
